from dataclasses import dataclass, field
from typing import Optional, Sequence, Union

import numpy as np
from numpy.typing import NDArray


@dataclass
class NHPPData:
    ages_at_events: NDArray[np.float64]
    events_assets_ids: Union[Sequence[str], NDArray[np.int64]]
    first_ages: Optional[NDArray[np.float64]] = field(repr=False, default=None)
    last_ages: Optional[NDArray[np.float64]] = field(repr=False, default=None)
    model_args: Optional[tuple[float | NDArray[np.float64], ...]] = field(repr=False, default=None)
    assets_ids: Optional[Union[Sequence[str], NDArray[np.int64]]] = field(repr=False, default=None)

    first_age_index: NDArray[np.int64] = field(repr=False, init=False)
    last_age_index: NDArray[np.int64] = field(repr=False, init=False)

    def __post_init__(self):
        # convert inputs to arrays
        self.events_assets_ids = np.unique(np.asarray(self.events_assets_ids), return_inverse=True)[1].astype(np.uint32)
        self.ages_at_events = np.asarray(self.ages_at_events, dtype=np.float64)
        if self.assets_ids is not None:
            self.assets_ids = np.unique(np.asarray(self.assets_ids), return_inverse=True)[1].astype(np.uint32)
        if self.first_ages is not None:
            self.first_ages = np.asarray(self.first_ages, dtype=np.float64)
        if self.last_ages is not None:
            self.last_ages = np.asarray(self.last_ages, dtype=np.float64)

        # control shapes
        if self.events_assets_ids.ndim != 1:
            raise ValueError("Invalid array shape for events_assets_ids. Expected 1d-array")
        if self.ages_at_events.ndim != 1:
            raise ValueError("Invalid array shape for ages. Expected 1d-array")
        if len(self.events_assets_ids) != len(self.ages_at_events):
            raise ValueError("Shape of events_assets_ids and ages must be equal. Expected equal length 1d-arrays")
        if self.assets_ids is not None:
            if self.assets_ids.ndim != 1:
                raise ValueError("Invalid array shape for assets_ids. Expected 1d-array")
            if self.first_ages is not None:
                if self.first_ages.ndim != 1:
                    raise ValueError("Invalid array shape for start_ages. Expected 1d-array")
                if len(self.first_ages) != len(self.assets_ids):
                    raise ValueError(
                        "Shape of assets_ids and start_ages must be equal. Expected equal length 1d-arrays"
                    )
            if self.last_ages is not None:
                if self.last_ages.ndim != 1:
                    raise ValueError("Invalid array shape for last_ages. Expected 1d-array")
                if len(self.last_ages) != len(self.assets_ids):
                    raise ValueError("Shape of assets_ids and last_ages must be equal. Expected equal length 1d-arrays")
            if bool(self.model_args):
                for arg in self.model_args:
                    arg = np.atleast_2d(np.asarray(arg, dtype=np.float64))
                    if arg.ndim > 2:
                        raise ValueError("Invalid arg shape in model_args. Arrays must be 0, 1 or 2d")
                    try:
                        arg.reshape((len(self.assets_ids), -1))
                    except ValueError:
                        raise ValueError(
                            "Invalid arg shape in model_args. Arrays must coherent with the number of assets given by assets_ids"
                        )
        else:
            if self.first_ages is not None:
                raise ValueError("If first_ages is given, corresponding asset ids must be given in assets_ids")
            if self.last_ages is not None:
                raise ValueError("If last_ages is given, corresponding asset ids must be given in assets_ids")
            if bool(self.model_args):
                raise ValueError("If model_args is given, corresponding asset ids must be given in assets_ids")

        # if self.events_assets_ids.dtype != np.int64:
        #     events_assets_ids = np.unique(self.events_assets_ids, return_inverse=True)[1]
        # # convert assets_id to int id
        # if self.assets_ids is not None:
        #     if self.assets_ids.dtype != np.int64:
        #         assets_ids = np.unique(self.assets_ids, return_inverse=True)[1]
        #     # control ids correspondance
        #     if not np.all(np.isin(self.events_assets_ids, self.assets_ids)):
        #         raise ValueError("If assets_ids is filled, all values of events_assets_ids must exist in assets_ids")

        # sort fields
        sort_ind = np.lexsort((self.ages_at_events, self.events_assets_ids))
        self.events_assets_ids = self.events_assets_ids[sort_ind]
        self.ages_at_events = self.ages_at_events[sort_ind]

        # number of age value per asset id
        nb_ages_per_asset = np.unique_counts(self.events_assets_ids).counts
        # index of the first ages and last ages in ages
        self.first_age_index = np.where(np.roll(self.events_assets_ids, 1) != self.events_assets_ids)[0]
        self.last_age_index = np.append(self.first_age_index[1:] - 1, len(self.events_assets_ids) - 1)

        if self.assets_ids is not None:

            # sort fields
            sort_ind = np.argsort(self.assets_ids)
            self.assets_ids = self.assets_ids[sort_ind]
            self.first_ages = self.first_ages[sort_ind] if self.first_ages is not None else self.first_ages
            self.last_ages = self.last_ages[sort_ind] if self.last_ages is not None else self.last_ages
            self.model_args = (
                tuple((arg[sort_ind] for arg in self.model_args)) if self.model_args is not None else self.model_args
            )

            if self.first_ages is not None:
                if np.any(self.ages_at_events[self.first_age_index] <= self.first_ages[nb_ages_per_asset != 0]):
                    raise ValueError("Each first_ages value must be lower than all of its corresponding ages values")
            if self.last_ages is not None:
                if np.any(self.ages_at_events[self.last_age_index] >= self.last_ages[nb_ages_per_asset != 0]):
                    raise ValueError("Each last_ages value must be greater than all of its corresponding ages values")

    def to_lifetime_data(self):
        event = np.ones_like(self.ages_at_events, dtype=np.bool_)
        # insert_index = np.cumsum(nb_ages_per_asset)
        # insert_index = last_age_index + 1
        if self.last_ages is not None:
            time = np.insert(self.ages_at_events, self.last_age_index + 1, self.last_ages)
            event = np.insert(event, self.last_age_index + 1, False)
            _ids = np.insert(self.events_assets_ids, self.last_age_index + 1, self.assets_ids)
            if self.first_ages is not None:
                entry = np.insert(
                    self.ages_at_events,
                    np.insert((self.last_age_index + 1)[:-1], 0, 0),
                    self.first_ages,
                )
            else:
                entry = np.insert(self.ages_at_events, self.first_age_index, 0.0)
        else:
            time = self.ages_at_events.copy()
            _ids = self.events_assets_ids.copy()
            if self.first_ages is not None:
                entry = np.roll(self.ages_at_events, 1)
                entry[self.first_age_index] = self.first_ages
            else:
                entry = np.roll(self.ages_at_events, 1)
                entry[self.first_age_index] = 0.0
        model_args = tuple((np.take(arg, _ids) for arg in self.model_args)) if self.model_args is not None else ()
        return time, event, entry, model_args



from .base import StochasticProcess
from .non_homogeneous_poisson_process import (
    FrozenNonHomogeneousPoissonProcess,
    NonHomogeneousPoissonProcess,
)
from .renewal_process import RenewalProcess, RenewalRewardProcess



import numpy as np
import pytest
from pytest import approx


def test_rvs(distribution, rvs_size, expected_out_shape):
    assert distribution.rvs(size=rvs_size).shape == expected_out_shape(size=rvs_size)
    assert all(
        arr.shape == expected_out_shape(size=rvs_size) for arr in distribution.rvs(size=rvs_size, return_event=True)
    )
    assert all(
        arr.shape == expected_out_shape(size=rvs_size) for arr in distribution.rvs(size=rvs_size, return_entry=True)
    )
    assert all(
        arr.shape == expected_out_shape(size=rvs_size)
        for arr in distribution.rvs(size=rvs_size, return_event=True, return_entry=True)
    )


def test_sf(distribution, time, expected_out_shape):
    assert distribution.sf(time).shape == expected_out_shape(time=time)
    assert distribution.sf(np.full(expected_out_shape(time=time), distribution.median())) == approx(
        np.full(expected_out_shape(time=time), 0.5), rel=1e-3
    )


def test_hf(distribution, time, expected_out_shape):
    assert distribution.hf(time).shape == expected_out_shape(time=time)


def test_chf(distribution, time, expected_out_shape):
    assert distribution.chf(time).shape == expected_out_shape(time=time)


def test_cdf(distribution, time, expected_out_shape):
    assert distribution.cdf(time).shape == expected_out_shape(time=time)


def test_pdf(distribution, time, expected_out_shape):
    assert distribution.pdf(time).shape == expected_out_shape(time=time)


def test_ppf(distribution, probability, expected_out_shape):
    assert distribution.ppf(probability).shape == expected_out_shape(probability=probability)


def test_ichf(distribution, probability, expected_out_shape):
    assert distribution.ichf(probability).shape == expected_out_shape(probability=probability)


def test_isf(distribution, probability, expected_out_shape):
    assert distribution.isf(probability).shape == expected_out_shape(probability=probability)
    assert distribution.isf(np.full(expected_out_shape(probability=probability), 0.5)) == approx(
        np.full(expected_out_shape(probability=probability), distribution.median())
    )


def test_moment(distribution, expected_out_shape):
    assert distribution.moment(1).shape == expected_out_shape()
    assert distribution.moment(2).shape == expected_out_shape()


def test_mean(distribution, expected_out_shape):
    assert distribution.mean().shape == expected_out_shape()


def test_var(distribution, expected_out_shape):
    assert distribution.var().shape == expected_out_shape()


def test_median(distribution, expected_out_shape):
    assert distribution.median().shape == expected_out_shape()


def test_dhf(distribution, time, expected_out_shape):
    assert distribution.dhf(time).shape == expected_out_shape(time=time)


def test_jac_sf(distribution, time, expected_out_shape):
    assert distribution.jac_sf(time, asarray=True).shape == (distribution.nb_params,) + expected_out_shape(time=time)


def test_jac_hf(distribution, time, expected_out_shape):
    assert distribution.jac_hf(time, asarray=True).shape == (distribution.nb_params,) + expected_out_shape(time=time)


def test_jac_chf(distribution, time, expected_out_shape):
    assert distribution.jac_chf(time, asarray=True).shape == (distribution.nb_params,) + expected_out_shape(time=time)


def test_jac_cdf(distribution, time, expected_out_shape):
    assert distribution.jac_cdf(time, asarray=True).shape == (distribution.nb_params,) + expected_out_shape(time=time)


def test_jac_pdf(distribution, time, expected_out_shape):
    assert distribution.jac_pdf(time, asarray=True).shape == (distribution.nb_params,) + expected_out_shape(time=time)


def test_ls_integrate(distribution, integration_bound_a, integration_bound_b, expected_out_shape):
    # integral_a^b dF(x)
    shape = expected_out_shape(integration_bound_a=integration_bound_a, integration_bound_b=integration_bound_b)
    integration = distribution.ls_integrate(np.ones_like, integration_bound_a, integration_bound_b, deg=100)
    assert integration.shape == shape
    assert integration == approx(distribution.cdf(integration_bound_b) - distribution.cdf(integration_bound_a))
    # integral_0^inf x*dF(x)
    integration = distribution.ls_integrate(
        lambda x: x, np.zeros_like(integration_bound_a), np.full_like(integration_bound_b, np.inf), deg=100
    )
    assert integration == approx(np.full(shape, distribution.mean()), rel=1e-3)


def test_fit(distribution, power_transformer_data):
    expected_params = distribution.params.copy()
    distribution = distribution.fit(
        power_transformer_data["time"],
        event=power_transformer_data["event"],
        entry=power_transformer_data["entry"],
    )
    assert distribution.params == pytest.approx(expected_params, rel=1e-3)


class TestEquilibriumDistribution:

    # def test_args_names(self, equilibrium_distribution):
    #     assert equilibrium_distribution.args_names == ()
    #
    # def test_rvs(self, equilibrium_distribution):
    #     m, n = 3, 10
    #     assert equilibrium_distribution.rvs(seed=21).shape == ()
    #     assert equilibrium_distribution.rvs(size=(n,), seed=21).shape == (n,)
    #     assert equilibrium_distribution.rvs(size=(m, 1), seed=21).shape == (m, 1)
    #     assert equilibrium_distribution.rvs(size=(m, n), seed=21).shape == (m, n)

    @pytest.mark.xfail
    def test_moment(self, equilibrium_distribution):
        assert equilibrium_distribution.moment(1).shape == ()
        assert equilibrium_distribution.moment(2).shape == ()

    @pytest.mark.xfail
    def test_mean(self, equilibrium_distribution):
        assert equilibrium_distribution.mean().shape == ()

    @pytest.mark.xfail
    def test_var(self, equilibrium_distribution):
        assert equilibrium_distribution.var().shape == ()

    @pytest.mark.xfail
    def test_median(self, equilibrium_distribution):
        assert equilibrium_distribution.median().shape == ()



from abc import ABC

import numpy as np
from scipy.optimize import Bounds, newton
from scipy.special import digamma, exp1, gamma, gammaincc, gammainccinv, polygamma

from relife.data import LifetimeData
from relife.likelihood import LikelihoodFromLifetimes
from relife.quadrature import laguerre_quadrature, legendre_quadrature

from ._base import FittableParametricLifetimeModel, ParametricLifetimeModel
from .regression import LifetimeRegression


class LifetimeDistribution(FittableParametricLifetimeModel, ABC):
    """
    Base class for distribution model.
    """

    def sf(self, time):
        """
        The survival function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return super().sf(time)

    def isf(self, probability):
        """
        The inverse survival function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).
        """
        cumulative_hazard_rate = -np.log(probability + 1e-6)  # avoid division by zero
        return self.ichf(cumulative_hazard_rate)

    def cdf(self, time):
        """
        The cumulative density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return super().cdf(time)

    def pdf(self, time):
        """
        The probability density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return super().pdf(time)

    def ppf(self, probability):
        """
        The percent point function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).
        """
        return super().ppf(probability)

    def moment(self, n):
        """
        n-th order moment

        Parameters
        ----------
        n : int
            order of the moment, at least 1.

        Returns
        -------
        np.float64
        """
        return super().moment(n)

    def median(self):
        """
        The median.

        Returns
        -------
        np.float64
        """
        return self.ppf(0.5)  # no super here to return np.float64

    def jac_sf(self, time, *, asarray=False):
        """
        The jacobian of the survival function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        jac_chf, sf = self.jac_chf(time, asarray=True), self.sf(time)
        jac = -jac_chf * sf
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_cdf(self, time, *, asarray=False):
        """
        The jacobian of the cumulative density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        jac = -self.jac_sf(time, asarray=True)
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_pdf(self, time, *, asarray=False):
        """
        The jacobian of the probability density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        jac_hf, hf = self.jac_hf(time, asarray=True), self.hf(time)
        jac_sf, sf = self.jac_sf(time, asarray=True), self.sf(time)
        jac = jac_hf * sf + jac_sf * hf
        if not asarray:
            return np.unstack(jac)
        return jac

    def rvs(self, size, *, nb_assets=None, return_event=False, return_entry=False, seed=None):
        """
        Random variable sampling.

        Parameters
        ----------
        size : int
            Size of the generated sample.
        nb_assets : int, optional
            If nb_assets is not None, 2d arrays of samples are generated.
        return_event : bool, default is False
            If True, returns event indicators along with the sample time values.
        return_entry : bool, default is False
            If True, returns corresponding entry values of the sample time values.
        seed : optional int, default is None
            Random seed used to fix random sampling.

        Returns
        -------
        float, ndarray or tuple of float or ndarray
            The sample values. If either ``return_event`` or ``return_entry`` is True, returns a tuple containing
            the time values followed by event values, entry values or both.
        """
        return super().rvs(size, nb_assets=nb_assets, return_event=return_event, return_entry=return_entry, seed=seed)

    def ls_integrate(self, func, a, b, deg=10):
        """
        Lebesgue-Stieltjes integration.

        Parameters
        ----------
        func : callable (in : 1 ndarray , out : 1 ndarray)
            The callable must have only one ndarray object as argument and one ndarray object as output
        a : ndarray (maximum number of dimension is 2)
            Lower bound(s) of integration.
        b : ndarray (maximum number of dimension is 2)
            Upper bound(s) of integration. If lower bound(s) is infinite, use np.inf as value.)
        deg : int, default 10
            Degree of the polynomials interpolation

        Returns
        -------
        np.ndarray
            Lebesgue-Stieltjes integral of func from `a` to `b`.
        """
        return super().ls_integrate(func, a, b, deg=deg)

    def _get_initial_params(self, time, event=None, entry=None, departure=None):
        param0 = np.ones(self.nb_params, dtype=np.float64)
        param0[-1] = 1 / np.median(time)
        self.params = param0
        return param0

    def _get_params_bounds(self):
        return Bounds(
            np.full(self.nb_params, np.finfo(float).resolution),
            np.full(self.nb_params, np.inf),
        )

    def fit(self, time, *, event=None, entry=None, departure=None, **options):
        """
        Estimation of the distribution parameters from lifetime data.

        Parameters
        ----------
        time : ndarray (1d or 2d)
            Observed lifetime values.
        event : ndarray of boolean values (1d), default is None
            Boolean indicators tagging lifetime values as right censored or complete.
        entry : ndarray of float (1d), default is None
            Left truncations applied to lifetime values.
        departure : ndarray of float (1d), default is None
            Right truncations applied to lifetime values.
        **options
            Extra arguments used by `scipy.minimize`. Default values are:
                - `method` : `"L-BFGS-B"`
                - `contraints` : `()`
                - `tol` : `None`
                - `callback` : `None`
                - `options` : `None`
                - `bounds` : `self.params_bounds`
                - `x0` : `self.init_params`

        Returns
        -------
        Self
            The current object with the estimated parameters setted inplace.

        Notes
        -----
        Supported lifetime observations format is either 1d-array or 2d-array. 2d-array is more advanced
        format that allows to pass other information as left-censored or interval-censored values. In this case,
        `event` is not needed as 2d-array encodes right-censored values by itself.
        """
        return super().fit(time, event=event, entry=entry, departure=departure, **options)


class Exponential(LifetimeDistribution):
    # noinspection PyUnresolvedReferences
    r"""
    Exponential lifetime distribution.

    The exponential distribution is a 1-parameter distribution with
    :math:`(\lambda)`. The probability density function is:

    .. math::

        f(t) = \lambda e^{-\lambda t}

    where:
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    rate : float, optional
        rate parameter

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    rate
    """

    def __init__(self, rate=None):
        super().__init__(rate=rate)

    @property
    def rate(self):  # optional but better for clarity and type checking
        """Get the current rate value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("rate")

    def hf(self, time):
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """

        return self.rate * np.ones_like(time)

    def chf(self, time):
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return np.asarray(self.rate) * time

    def mean(self):
        """
        The mean of the distribution.

        Returns
        -------
        np.float64
        """
        return 1 / np.asarray(self.rate)

    def var(self):
        """
        The variance of the distribution.

        Returns
        -------
        np.float64
        """
        return 1 / np.asarray(self.rate) ** 2

    def mrl(self, time):
        """
        The mean residual life function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return 1 / self.rate * np.ones_like(time)

    def ichf(self, cumulative_hazard_rate):
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are (), (n,) or (m, n).

        Returns
        -------
            Function values at each given cumulative hazard rate(s).
        """
        return cumulative_hazard_rate / np.asarray(self.rate)

    def jac_hf(self, time, *, asarray=False):
        """
        The jacobian of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        if isinstance(time, np.ndarray):
            jac = np.expand_dims(np.ones_like(time), axis=0).copy()
        else:
            jac = np.array([1], dtype=np.float64)
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_chf(self, time, *, asarray=False):
        """
        The jacobian of the cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        if isinstance(time, np.ndarray):
            jac = np.expand_dims(time, axis=0).copy()
        else:
            jac = np.array([time], dtype=np.float64)
        if not asarray:
            return np.unstack(jac)
        return jac

    def dhf(self, time):
        """
        The derivate of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        if isinstance(time, np.ndarray):
            return np.zeros_like(time)
        return np.asarray(0.0)


class Weibull(LifetimeDistribution):
    # noinspection PyUnresolvedReferences
    r"""
    Weibull lifetime distribution.

    The Weibull distribution is a 2-parameter distribution with
    :math:`(c,\lambda)`. The probability density function is:

    .. math::

        f(t) = c \lambda (\lambda t)^{c-1} e^{-(\lambda t)^c}

    where:
        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    Parameters
    ----------
    shape : float, optional
        shape parameter
    rate : float, optional
        rate parameter

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape=None, rate=None):
        super().__init__(shape=shape, rate=rate)

    @property
    def shape(self):  # optional but better for clarity and type checking
        """Get the current shape value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("shape")

    @property
    def rate(self):  # optional but better for clarity and type checking
        """Get the current rate value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("rate")

    def hf(self, time):
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.shape * self.rate * (self.rate * np.asarray(time)) ** (self.shape - 1)

    def chf(self, time):
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return (self.rate * np.asarray(time)) ** self.shape

    def mean(self):
        """
        The mean of the distribution.

        Returns
        -------
        np.float64
        """
        return gamma(1 + 1 / self.shape) / self.rate

    def var(self):
        """
        The variance of the distribution.

        Returns
        -------
        np.float64
        """
        return gamma(1 + 2 / self.shape) / self.rate**2 - self.mean() ** 2

    def mrl(self, time):
        """
        The mean residual life function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return (
            gamma(1 / self.shape)
            / (self.rate * self.shape * self.sf(time))
            * gammaincc(
                1 / self.shape,
                (self.rate * time) ** self.shape,
            )
        )

    def ichf(self, cumulative_hazard_rate):
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are (), (n,) or (m, n).

        Returns
        -------
            Function values at each given cumulative hazard rate(s).
        """
        return np.asarray(cumulative_hazard_rate) ** (1 / self.shape) / self.rate

    def jac_hf(self, time, *, asarray=False):
        """
        The jacobian of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        jac = (
            self.rate * (self.rate * time) ** (self.shape - 1) * (1 + self.shape * np.log(self.rate * time)),
            self.shape**2 * (self.rate * time) ** (self.shape - 1),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def jac_chf(self, time, *, asarray=False):
        """
        The jacobian of the cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        jac = (
            np.log(self.rate * time) * (self.rate * time) ** self.shape,
            self.shape * time * (self.rate * time) ** (self.shape - 1),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def dhf(self, time):
        """
        The derivate of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        time = np.asarray(time)
        return self.shape * (self.shape - 1) * self.rate**2 * (self.rate * time) ** (self.shape - 2)


class Gompertz(LifetimeDistribution):
    # noinspection PyUnresolvedReferences
    r"""
    Gompertz lifetime distribution.

    The Gompertz distribution is a 2-parameter distribution with
    :math:`(c,\lambda)`. The probability density function is:

    .. math::

        f(t) = c \lambda e^{\lambda t} e^{ -c \left( e^{\lambda t}-1 \right) }

    where:

        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    shape : float, optional
        shape parameter
    rate : float, optional
        rate parameter

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape=None, rate=None):
        super().__init__(shape=shape, rate=rate)

    @property
    def shape(self):  # optional but better for clarity and type checking
        """Get the current shape value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("shape")

    @property
    def rate(self):  # optional but better for clarity and type checking
        """Get the current rate value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("rate")

    def hf(self, time):
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.shape * self.rate * np.exp(self.rate * time)

    def chf(self, time):
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.shape * np.expm1(self.rate * time)

    def mean(self):
        """
        The mean of the distribution.

        Returns
        -------
        np.float64
        """
        return np.exp(self.shape) * exp1(self.shape) / self.rate

    def var(self):
        """
        The variance of the distribution.

        Returns
        -------
        np.float64
        """
        return polygamma(1, 1) / self.rate**2

    def mrl(self, time):
        """
        The mean residual life function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        z = self.shape * np.exp(self.rate * time)
        return np.exp(z) * exp1(z) / self.rate

    def ichf(self, cumulative_hazard_rate):
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are (), (n,) or (m, n).

        Returns
        -------
            Function values at each given cumulative hazard rate(s).
        """
        return 1 / self.rate * np.log1p(cumulative_hazard_rate / self.shape)

    def jac_hf(self, time, *, asarray=False):
        """
        The jacobian of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        jac = (
            self.rate * np.exp(self.rate * time),
            self.shape * np.exp(self.rate * time) * (1 + self.rate * time),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def jac_chf(self, time, *, asarray=False):
        """
        The jacobian of the cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        jac = (
            np.expm1(self.rate * time),
            self.shape * time * np.exp(self.rate * time),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def dhf(self, time):
        """
        The derivate of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.shape * self.rate**2 * np.exp(self.rate * time)

    def _get_initial_params(self, time, event=None, entry=None, departure=None):
        param0 = np.empty(self.nb_params, dtype=np.float64)
        rate = np.pi / (np.sqrt(6) * np.std(time))
        shape = np.exp(-rate * np.mean(time))
        param0[0] = shape
        param0[1] = rate
        return param0


class Gamma(LifetimeDistribution):
    # noinspection PyUnresolvedReferences
    r"""
    Gamma lifetime distribution.

    The Gamma distribution is a 2-parameter distribution with
    :math:`(c,\lambda)`. The probability density function is:

    .. math::

        f(t) = \frac{\lambda^c t^{c-1} e^{-\lambda t}}{\Gamma(c)}

    where:

        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    shape : float, optional
        shape parameter
    rate : float, optional
        rate parameter

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape=None, rate=None):
        super().__init__(shape=shape, rate=rate)

    def _uppergamma(self, x):
        return gammaincc(self.shape, x) * gamma(self.shape)

    def _jac_uppergamma_shape(self, x):
        return laguerre_quadrature(lambda s: np.log(s) * s ** (self.shape - 1), x, deg=100)

    @property
    def shape(self):  # optional but better for clarity and type checking
        """Get the current shape value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("shape")

    @property
    def rate(self):  # optional but better for clarity and type checking
        """Get the current rate value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("rate")

    def hf(self, time):
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        x = self.rate * time
        return self.rate * x ** (self.shape - 1) * np.exp(-x) / self._uppergamma(x)

    def chf(self, time):
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        x = self.rate * time
        return np.log(gamma(self.shape)) - np.log(self._uppergamma(x))

    def mean(self):
        """
        The mean of the distribution.

        Returns
        -------
        np.float64
        """
        return np.asarray(self.shape / self.rate)

    def var(self):
        """
        The variance of the distribution.

        Returns
        -------
        np.float64
        """
        return np.asarray(self.shape / (self.rate**2))

    def ichf(self, cumulative_hazard_rate):
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are (), (n,) or (m, n).

        Returns
        -------
            Function values at each given cumulative hazard rate(s).
        """
        return 1 / self.rate * gammainccinv(self.shape, np.exp(-cumulative_hazard_rate))

    def jac_hf(self, time, *, asarray=False):
        """
        The jacobian of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        x = self.rate * time
        y = x ** (self.shape - 1) * np.exp(-x) / self._uppergamma(x) ** 2
        jac = (
            y * ((self.rate * np.log(x) * self._uppergamma(x)) - self.rate * self._jac_uppergamma_shape(x)),
            y * ((self.shape - x) * self._uppergamma(x) + x**self.shape * np.exp(-x)),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def jac_chf(self, time, *, asarray=False):
        """
        The jacobian of the cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        x = self.rate * time
        jac = (
            digamma(self.shape) - self._jac_uppergamma_shape(x) / self._uppergamma(x),
            (x ** (self.shape - 1) * time * np.exp(-x) / self._uppergamma(x)),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def dhf(self, time):
        """
        The derivate of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.hf(time) * ((self.shape - 1) / time - self.rate + self.hf(time))

    def mrl(self, time):
        """
        The mean residual life function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return super().mrl(time)


class LogLogistic(LifetimeDistribution):
    # noinspection PyUnresolvedReferences
    r"""
    Log-logistic probability distribution.

    The Log-logistic distribution is defined as a 2-parameter distribution
    :math:`(c, \lambda)`. The probability density function is:

    .. math::

        f(t) = \frac{c \lambda^c t^{c-1}}{(1+(\lambda t)^{c})^2}

    where:

        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    shape : float, optional
        shape parameter
    rate : float, optional
        rate parameter

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape=None, rate=None):
        super().__init__(shape=shape, rate=rate)

    @property
    def shape(self):  # optional but better for clarity and type checking
        """Get the current shape value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("shape")

    @property
    def rate(self):  # optional but better for clarity and type checking
        """Get the current rate value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("rate")

    def hf(self, time):
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        x = self.rate * np.asarray(time)
        return self.shape * self.rate * x ** (self.shape - 1) / (1 + x**self.shape)

    def chf(self, time):
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        x = self.rate * time
        return np.log(1 + x**self.shape)

    def mean(self):
        """
        The mean of the distribution.

        Returns
        -------
        np.float64
        """
        b = np.pi / self.shape
        if self.shape <= 1:
            raise ValueError(f"Expectancy only defined for shape > 1: shape = {self.shape}")
        return b / (self.rate * np.sin(b))

    def var(self):
        """
        The variance of the distribution.

        Returns
        -------
        np.float64
        """
        b = np.pi / self.shape
        if self.shape <= 2:
            raise ValueError(f"Variance only defined for shape > 2: shape = {self.shape}")
        return (1 / self.rate**2) * (2 * b / np.sin(2 * b) - b**2 / (np.sin(b) ** 2))

    def ichf(self, cumulative_hazard_rate):
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are (), (n,) or (m, n).

        Returns
        -------
            Function values at each given cumulative hazard rate(s).
        """
        return ((np.exp(cumulative_hazard_rate) - 1) ** (1 / self.shape)) / self.rate

    def jac_hf(self, time, *, asarray=False):
        """
        The jacobian of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        x = self.rate * time
        jac = (
            (self.rate * x ** (self.shape - 1) / (1 + x**self.shape) ** 2)
            * (1 + x**self.shape + self.shape * np.log(self.rate * time)),
            (self.rate * x ** (self.shape - 1) / (1 + x**self.shape) ** 2) * (self.shape**2 / self.rate),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def jac_chf(self, time, *, asarray=False):
        """
        The jacobian of the cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        x = self.rate * time
        jac = (
            (x**self.shape / (1 + x**self.shape)) * np.log(self.rate * time),
            (x**self.shape / (1 + x**self.shape)) * (self.shape / self.rate),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def dhf(self, time):
        """
        The derivate of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        x = self.rate * np.asarray(time)
        return (
            self.shape
            * self.rate**2
            * x ** (self.shape - 2)
            * (self.shape - 1 - x**self.shape)
            / (1 + x**self.shape) ** 2
        )

    def mrl(self, time):
        return super().mrl(time)


class EquilibriumDistribution(ParametricLifetimeModel):
    r"""Equilibrium distribution.

    The equilibirum distribution is the distrbution computed from a lifetime
    core that makes the associated delayed renewal stochastic_process stationary.

    Parameters
    ----------
    baseline : BaseLifetimeModel
        Underlying lifetime core.

    References
    ----------
    .. [1] Ross, S. M. (1996). Stochastic stochastic_process. New York: Wiley.
    """

    # can't expect baseline to be FrozenParametricLifetimeModel too because it does not have freeze_args
    def __init__(self, baseline):
        super().__init__()
        self.baseline = baseline

    def cdf(self, time, *args):
        return legendre_quadrature(lambda x: self.baseline.sf(x, *args), 0, time) / self.baseline.mean(*args)

    def sf(self, time, *args):
        return 1 - self.cdf(time, *args)

    def pdf(self, time, *args):
        return self.baseline.sf(time, *args) / self.baseline.mean(*args)

    def hf(self, time, *args):
        return 1 / self.baseline.mrl(time, *args)

    def chf(self, time, *args):
        return -np.log(self.sf(time, *args))

    def isf(self, probability, *args):
        return newton(
            lambda x: self.sf(x, *args) - probability,
            self.baseline.isf(probability, *args),
            args=args,
        )

    def ichf(
        self,
        cumulative_hazard_rate,
        *args,
    ):
        return self.isf(np.exp(-cumulative_hazard_rate), *args)


class MinimumDistribution(ParametricLifetimeModel):
    r"""Series structure of n identical and independent components.

    The hazard function of the system is given by:

    .. math::

        h(t) = n \cdot  h_0(t)

    where :math:`h_0` is the baseline hazard function of the components.

    Examples
    --------

    Computing the survival (or reliability) function for 3 structures of 3,6 and
    9 identical and idependent components:

    .. code-block::

        model = MinimumDistribution(Weibull(2, 0.05))
        t = np.arange(0, 10, 0.1)
        n = np.array([3, 6, 9]).reshape(-1, 1)
        model.sf(t, n)
    """

    def __init__(self, baseline):
        super().__init__()
        self.baseline = baseline
        self.fitting_results = None

    def sf(self, time, n, *args):
        return super().sf(time)

    def pdf(self, time, n, *args):
        return super().pdf(time)

    def hf(self, time, n, *args):
        return n * self.baseline.hf(time, *args)

    def chf(self, time, n, *args):
        return n * self.baseline.chf(time, *args)

    def ichf(
        self,
        cumulative_hazard_rate,
        n,
        *args,
    ):
        return self.baseline.ichf(cumulative_hazard_rate / n, *args)

    def dhf(self, time, n, *args):
        return n * self.baseline.dhf(time, *args)

    def jac_chf(self, time, n, *args, asarray=False):
        return n * self.baseline.jac_chf(time, *args, asarray=asarray)

    def jac_hf(self, time, n, *args, asarray=False):
        return n * self.baseline.jac_hf(time, *args, asarray=asarray)

    def jac_sf(self, time, n, *args, asarray=False):
        jac_chf, sf = self.jac_chf(time, n, *args, asarray=True), self.sf(time)
        jac = -jac_chf * sf
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_cdf(self, time, n, *args, asarray=False):
        jac = -self.jac_sf(time, n, *args, asarray=True)
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_pdf(self, time, n, *args, asarray=False):
        jac_hf, hf = self.jac_hf(time, n, *args, asarray=True), self.hf(time, n)
        jac_sf, sf = self.jac_sf(time, n, *args, asarray=True), self.sf(time, n)
        jac = jac_hf * sf + jac_sf * hf
        if not asarray:
            return np.unstack(jac)
        return jac

    def ls_integrate(self, func, a, b, n, *args, deg: int = 10):
        return super().ls_integrate(func, a, b, n, *args, deg=deg)

    def fit(
        self,
        time,
        n,
        *args,
        event=None,
        entry=None,
        departure=None,
        **kwargs,
    ):
        # initialize params structure (number of parameters in params tree)
        if isinstance(self.baseline, LifetimeRegression):
            self.baseline._init_params(args[0], *args[1:])
        lifetime_data: LifetimeData = LifetimeData(time, event=event, entry=entry, departure=departure, args=(n, *args))
        likelihood = LikelihoodFromLifetimes(self, lifetime_data)
        fitting_results = likelihood.maximum_likelihood_estimation(**kwargs)
        self.params = fitting_results.optimal_params
        self.fitting_results = fitting_results
        return self



from relife import FrozenParametricModel, ParametricModel


class StochasticProcess(ParametricModel): ...


class FrozenStochasticProcess(FrozenParametricModel):
    def __init__(self, model: StochasticProcess, *args):
        super().__init__(model, *args)



from ._base import BaseAgeReplacementPolicy, BaseOneCycleAgeReplacementPolicy
from .age_replacement_policy import (
    AgeReplacementPolicy,
    NonHomogeneousPoissonAgeReplacementPolicy,
    OneCycleAgeReplacementPolicy,
)
from .run_to_failure_policy import OneCycleRunToFailurePolicy, RunToFailurePolicy



import numpy as np


def test_structured_lifetime_data_with_1d(lifetime_data_1d):
    assert np.all(lifetime_data_1d.complete.lifetime_index == np.array([0, 2, 6]).astype(np.int64))
    assert np.all(lifetime_data_1d.complete.lifetime_values == np.array([10, 9, 11]).astype(np.float64).reshape(-1, 1))
    assert np.all(lifetime_data_1d.right_censoring.lifetime_index == np.array([1, 3, 4, 5]))
    assert np.all(
        lifetime_data_1d.right_censoring.lifetime_values == np.array([11, 10, 12, 13]).astype(np.float64).reshape(-1, 1)
    )
    assert np.all(lifetime_data_1d.left_truncation.lifetime_index == np.array([2, 3, 4, 5, 6]).astype(np.int64))
    assert np.all(
        lifetime_data_1d.left_truncation.lifetime_values == np.array([3, 5, 3, 1, 9]).astype(np.float64).reshape(-1, 1)
    )


def test_structured_lifetime_data_with_2d(lifetime_data_2d):
    assert np.all(lifetime_data_2d.left_censoring.lifetime_index == np.array([1]).astype(np.int64))

    assert np.all(lifetime_data_2d.left_censoring.lifetime_values == np.array([4]).astype(np.float64).reshape(-1, 1))

    assert np.all(lifetime_data_2d.right_censoring.lifetime_index == np.array([3]))
    assert np.all(lifetime_data_2d.right_censoring.lifetime_values == np.array([7]).astype(np.float64).reshape(-1, 1))

    assert np.all(lifetime_data_2d.interval_censoring.lifetime_index == np.array([0, 1, 3, 5, 6]).astype(np.int64))
    assert np.all(
        lifetime_data_2d.interval_censoring.lifetime_values
        == np.array([[1, 2], [0, 4], [7, np.inf], [2, 10], [10, 11]]).astype(np.float64)
    )

    assert np.all(lifetime_data_2d.left_truncation.lifetime_values == np.array([3, 5, 3, 1, 9]).reshape(-1, 1))



from typing import Optional, TypeAlias, TypedDict, TypeVar

import numpy as np
from numpy.typing import NDArray as NDArray
from typing_extensions import override

from relife._typing import _ParametricLifetimeModel
from relife.economic import ExponentialDiscounting, Reward

from ._sample import RenewalProcessSample, RenewalRewardProcessSample
from .base import StochasticProcess as StochasticProcess

_N = TypeVar("_N", bound=int)
_M: TypeAlias = int
_TL: TypeAlias = np.ndarray[tuple[_N], np.dtype[np.float64]] | np.ndarray[tuple[_M, _N], np.dtype[np.float64]]
_C: TypeAlias = np.ndarray[tuple[_N], np.dtype[np.float64]] | np.ndarray[tuple[_M, _N], np.dtype[np.float64]]
_A_C: TypeAlias = np.float64 | np.ndarray[tuple[_M], np.dtype[np.float64]]

class LifetimeFitArg(TypedDict):
    time: NDArray[np.float64]
    event: NDArray[np.bool_]
    entry: NDArray[np.float64]
    args: tuple[NDArray[np.float64], ...]

class RenewalProcess(StochasticProcess[()]):
    lifetime_model: _ParametricLifetimeModel[()]
    first_lifetime_model: Optional[_ParametricLifetimeModel[()]]
    def __init__(
        self,
        lifetime_model: _ParametricLifetimeModel[()],
        first_lifetime_model: Optional[_ParametricLifetimeModel[()]] = None,
    ) -> None: ...
    def _make_timeline(self, tf: float, nb_steps: _N) -> _TL[_N]: ...
    def renewal_function(self, tf: float, nb_steps: _N) -> tuple[_TL[_N], _C[_N]]: ...
    def renewal_density(self, tf: float, nb_steps: _N) -> tuple[_TL[_N], _C[_N]]: ...
    def sample(self, size: int, tf: float, t0: float = 0.0, seed: Optional[int] = None) -> RenewalProcessSample: ...
    def generate_failure_data(
        self, size: int, tf: float, t0: float = 0.0, seed: Optional[int] = None
    ) -> LifetimeFitArg: ...

class RenewalRewardProcess(RenewalProcess):
    reward: Reward
    first_reward: Optional[Reward]
    discounting: ExponentialDiscounting
    def __init__(
        self,
        lifetime_model: _ParametricLifetimeModel[()],
        reward: Reward,
        discounting_rate: float = 0.0,
        first_lifetime_model: Optional[_ParametricLifetimeModel[()]] = None,
        first_reward: Optional[Reward] = None,
    ) -> None: ...
    @property
    def discounting_rate(self) -> float: ...
    # noinspection PyUnresolvedReferences
    @discounting_rate.setter
    def discounting_rate(self, value: float) -> None: ...
    @override
    def _make_timeline(self, tf: float, nb_steps: _N) -> _TL[_N]: ...
    def expected_total_reward(self, tf: float, nb_steps: _N) -> tuple[_TL[_N], _C[_N]]: ...
    def expected_equivalent_annual_worth(self, tf: float, nb_steps: int) -> tuple[_TL[_N], _C[_N]]: ...
    def asymptotic_expected_total_reward(self) -> _A_C: ...
    def asymptotic_expected_equivalent_annual_worth(self) -> _A_C: ...
    @override
    def sample(
        self, size: int, tf: float, t0: float = 0.0, seed: Optional[int] = None
    ) -> RenewalRewardProcessSample: ...



from __future__ import annotations

from dataclasses import dataclass, replace
from typing import Optional

import numpy as np
from numpy.typing import NDArray


@dataclass
class CountDataSample:
    t0: float
    tf: float
    struct_array: NDArray[np.void]

    def select(self, sample_id: Optional[int] = None, asset_id: Optional[int] = None) -> CountDataSample:
        mask: NDArray[np.bool_] = np.ones_like(self.struct_array, dtype=np.bool_)
        if sample_id is not None:
            mask = mask & np.isin(self.struct_array["sample_id"], sample_id)
        if asset_id is not None:
            mask = mask & np.isin(self.struct_array["asset_id"], asset_id)
        struct_subarray = self.struct_array[mask].copy()
        return replace(self, t0=self.t0, tf=self.tf, struct_array=struct_subarray)

    @property
    def time(self) -> NDArray[np.float64]:
        return self.struct_array["time"]

    @property
    def timeline(self) -> NDArray[np.float64]:
        return self.struct_array["timeline"]

    @property
    def sample_id(self) -> NDArray[np.uint32]:
        return self.struct_array["sample_id"]

    @property
    def asset_id(self) -> NDArray[np.uint32]:
        return self.struct_array["asset_id"]

    @property
    def nb_renewal(self) -> NDArray[np.uint32]:
        return self.struct_array["nb_renewal"]

    @property
    def event(self) -> NDArray[np.bool_]:
        return self.struct_array["event"]

    @property
    def entry(self) -> NDArray[np.float64]:
        return self.struct_array["entry"]


@dataclass
class RenewalProcessSample(CountDataSample):

    @staticmethod
    def _nb_events(selection: RenewalProcessSample) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        sort = np.argsort(selection.struct_array["timeline"])
        timeline = selection.struct_array["timeline"][sort]
        counts = selection.struct_array["event"][sort]

        timeline = np.insert(timeline, 0, selection.t0)
        counts = np.insert(counts, 0, 0)
        counts[timeline == selection.tf] = 0
        return timeline, np.cumsum(counts)

    def nb_events(self, sample_id: int, asset_id: int = 0) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        selection = self.select(sample_id=sample_id, asset_id=asset_id)
        return RenewalProcessSample._nb_events(selection)

    def mean_nb_events(self, asset_id: int = 0) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        selection = self.select(asset_id=asset_id)
        timeline, counts = RenewalProcessSample._nb_events(selection)
        nb_sample = len(np.unique(selection.struct_array["sample_id"]))
        return timeline, counts / nb_sample


@dataclass
class RenewalRewardProcessSample(RenewalProcessSample):

    @staticmethod
    def _total_rewards(selection: RenewalRewardProcessSample) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        sort = np.argsort(selection.struct_array["timeline"])
        timeline = selection.struct_array["timeline"][sort]
        reward = selection.struct_array["reward"][sort]

        timeline = np.insert(timeline, 0, selection.t0)
        reward = np.insert(reward, 0, 0)
        reward[timeline == selection.tf] = 0

        return timeline, np.cumsum(reward)

    def total_rewards(self, sample_id: int, asset_id: int = 0) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        selection = self.select(sample_id=sample_id, asset_id=asset_id)
        return RenewalRewardProcessSample._total_rewards(selection)

    def mean_total_rewards(self, asset_id: int = 0) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        selection = self.select(asset_id=asset_id)
        timeline, rewards = RenewalRewardProcessSample._nb_events(selection)
        nb_sample = len(np.unique(selection.struct_array["sample_id"]))
        return timeline, rewards / nb_sample


NonHomogeneousPoissonProcessSample = RenewalProcessSample



import numpy as np
from pytest import approx

from relife.lifetime_model import AgeReplacementModel, LeftTruncatedModel


class TestAgeReplacementDistribution:

    def test_rvs(self, distribution, rvs_size, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).rvs(rvs_size, ar).shape == expected_out_shape(ar=ar, size=rvs_size)
        assert all(
            arr.shape == expected_out_shape(ar=ar, size=rvs_size)
            for arr in AgeReplacementModel(distribution).rvs(rvs_size, ar, return_event=True)
        )
        assert all(
            arr.shape == expected_out_shape(ar=ar, size=rvs_size)
            for arr in AgeReplacementModel(distribution).rvs(rvs_size, ar, return_entry=True)
        )
        assert all(
            arr.shape == expected_out_shape(ar=ar, size=rvs_size)
            for arr in AgeReplacementModel(distribution).rvs(rvs_size, ar, return_event=True, return_entry=True)
        )

    def test_sf(self, distribution, time, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).sf(time, ar).shape == expected_out_shape(time=time, ar=ar)

    def test_hf(self, distribution, time, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).hf(time, ar).shape == expected_out_shape(time=time, ar=ar)

    def test_chf(self, distribution, time, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).chf(time, ar).shape == expected_out_shape(time=time, ar=ar)

    def test_cdf(self, distribution, time, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).cdf(time, ar).shape == expected_out_shape(time=time, ar=ar)

    def test_pdf(self, distribution, time, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).pdf(time, ar).shape == expected_out_shape(time=time, ar=ar)

    def test_ppf(self, distribution, time, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).ppf(time, ar).shape == expected_out_shape(time=time, ar=ar)

    def test_ichf(self, distribution, probability, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).ichf(probability, ar).shape == expected_out_shape(
            probability=probability, ar=ar
        )

    def test_isf(self, distribution, probability, ar, expected_out_shape):
        out_shape = expected_out_shape(probability=probability, ar=ar)
        assert AgeReplacementModel(distribution).isf(probability, ar).shape == out_shape
        assert AgeReplacementModel(distribution).isf(np.full(out_shape, 0.5), ar) == approx(
            np.broadcast_to(AgeReplacementModel(distribution).median(ar), out_shape)
        )

    def test_moment(self, distribution, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).moment(1, ar).shape == expected_out_shape(ar=ar)
        assert AgeReplacementModel(distribution).moment(2, ar).shape == expected_out_shape(ar=ar)

    def test_mean(self, distribution, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).mean(ar).shape == expected_out_shape(ar=ar)

    def test_var(self, distribution, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).var(ar).shape == expected_out_shape(ar=ar)

    def test_median(self, distribution, ar, expected_out_shape):
        assert AgeReplacementModel(distribution).median(ar).shape == expected_out_shape(ar=ar)

    def test_ls_integrate(self, distribution, integration_bound_a, integration_bound_b, ar, expected_out_shape):
        np.random.seed(10)
        ar = np.random.uniform(2.0, 8.0, size=ar.shape)
        # integral_a^b dF(x)
        out_shape = expected_out_shape(
            integration_bound_a=integration_bound_a, integration_bound_b=integration_bound_b, ar=ar
        )
        integration = AgeReplacementModel(distribution).ls_integrate(
            np.ones_like, integration_bound_a, integration_bound_b, ar, deg=100
        )
        assert integration.shape == out_shape
        assert integration == approx(
            AgeReplacementModel(distribution).cdf(integration_bound_b, ar)
            - AgeReplacementModel(distribution).cdf(integration_bound_a, ar)
        )
        # integral_0^inf x*dF(x)
        integration = AgeReplacementModel(distribution).ls_integrate(
            lambda x: x, np.zeros_like(integration_bound_a), np.full_like(integration_bound_b, np.inf), ar, deg=100
        )
        assert integration == approx(np.broadcast_to(AgeReplacementModel(distribution).mean(ar), out_shape), rel=1e-3)


class TestAgeReplacementRegression:

    def test_rvs(self, regression, rvs_size, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).rvs(rvs_size, ar, covar).shape == expected_out_shape(
            ar=ar, covar=covar, size=rvs_size
        )
        assert all(
            arr.shape == expected_out_shape(ar=ar, covar=covar, size=rvs_size)
            for arr in AgeReplacementModel(regression).rvs(rvs_size, ar, covar, return_event=True)
        )
        assert all(
            arr.shape == expected_out_shape(ar=ar, covar=covar, size=rvs_size)
            for arr in AgeReplacementModel(regression).rvs(rvs_size, ar, covar, return_entry=True)
        )
        assert all(
            arr.shape == expected_out_shape(ar=ar, covar=covar, size=rvs_size)
            for arr in AgeReplacementModel(regression).rvs(rvs_size, ar, covar, return_event=True, return_entry=True)
        )

    def test_sf(self, regression, time, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).sf(time, ar, covar).shape == expected_out_shape(
            time=time, ar=ar, covar=covar
        )

    def test_hf(self, regression, time, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).hf(time, ar, covar).shape == expected_out_shape(
            time=time, ar=ar, covar=covar
        )

    def test_chf(self, regression, time, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).chf(time, ar, covar).shape == expected_out_shape(
            time=time, ar=ar, covar=covar
        )

    def test_cdf(self, regression, time, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).cdf(time, ar, covar).shape == expected_out_shape(
            time=time, ar=ar, covar=covar
        )

    def test_pdf(self, regression, time, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).pdf(time, ar, covar).shape == expected_out_shape(
            time=time, ar=ar, covar=covar
        )

    def test_ppf(self, regression, time, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).ppf(time, ar, covar).shape == expected_out_shape(
            time=time, ar=ar, covar=covar
        )

    def test_ichf(self, regression, probability, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).ichf(probability, ar, covar).shape == expected_out_shape(
            probability=probability, ar=ar, covar=covar
        )

    def test_isf(self, regression, probability, ar, covar, expected_out_shape):
        out_shape = expected_out_shape(probability=probability, ar=ar, covar=covar)
        assert AgeReplacementModel(regression).isf(probability, ar, covar).shape == out_shape
        assert AgeReplacementModel(regression).isf(np.full(out_shape, 0.5), ar, covar) == approx(
            np.broadcast_to(AgeReplacementModel(regression).median(ar, covar), out_shape)
        )

    def test_moment(self, regression, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).moment(1, ar, covar).shape == expected_out_shape(ar=ar, covar=covar)
        assert AgeReplacementModel(regression).moment(2, ar, covar).shape == expected_out_shape(ar=ar, covar=covar)

    def test_mean(self, regression, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).mean(ar, covar).shape == expected_out_shape(ar=ar, covar=covar)

    def test_var(self, regression, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).var(ar, covar).shape == expected_out_shape(ar=ar, covar=covar)

    def test_median(self, regression, ar, covar, expected_out_shape):
        assert AgeReplacementModel(regression).median(ar, covar).shape == expected_out_shape(ar=ar, covar=covar)

    def test_ls_integrate(self, regression, integration_bound_a, integration_bound_b, ar, covar, expected_out_shape):
        np.random.seed(10)
        ar = np.random.uniform(2.0, 8.0, size=ar.shape)
        # integral_a^b dF(x)
        out_shape = expected_out_shape(
            integration_bound_a=integration_bound_a, integration_bound_b=integration_bound_b, ar=ar, covar=covar
        )
        integration = AgeReplacementModel(regression).ls_integrate(
            np.ones_like, integration_bound_a, integration_bound_b, ar, covar, deg=100
        )
        assert integration.shape == out_shape
        assert integration == approx(
            AgeReplacementModel(regression).cdf(integration_bound_b, ar, covar)
            - AgeReplacementModel(regression).cdf(integration_bound_a, ar, covar)
        )
        # integral_0^inf x*dF(x)
        integration = AgeReplacementModel(regression).ls_integrate(
            lambda x: x,
            np.zeros_like(integration_bound_a),
            np.full_like(integration_bound_b, np.inf),
            ar,
            covar,
            deg=100,
        )
        assert integration == approx(
            np.broadcast_to(AgeReplacementModel(regression).mean(ar, covar), out_shape), rel=1e-3
        )


class TestLeftTruncatedDistribution:

    def test_rvs(self, distribution, rvs_size, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).rvs(rvs_size, a0).shape == expected_out_shape(a0=a0, size=rvs_size)
        assert all(
            arr.shape == expected_out_shape(a0=a0, size=rvs_size)
            for arr in LeftTruncatedModel(distribution).rvs(rvs_size, a0, return_event=True)
        )
        assert all(
            arr.shape == expected_out_shape(a0=a0, size=rvs_size)
            for arr in LeftTruncatedModel(distribution).rvs(rvs_size, a0, return_entry=True)
        )
        assert all(
            arr.shape == expected_out_shape(a0=a0, size=rvs_size)
            for arr in LeftTruncatedModel(distribution).rvs(rvs_size, a0, return_event=True, return_entry=True)
        )

    def test_sf(self, distribution, time, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).sf(time, a0).shape == expected_out_shape(time=time, a0=a0)

    def test_hf(self, distribution, time, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).hf(time, a0).shape == expected_out_shape(time=time, a0=a0)

    def test_chf(self, distribution, time, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).chf(time, a0).shape == expected_out_shape(time=time, a0=a0)

    def test_cdf(self, distribution, time, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).cdf(time, a0).shape == expected_out_shape(time=time, a0=a0)

    def test_pdf(self, distribution, time, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).pdf(time, a0).shape == expected_out_shape(time=time, a0=a0)

    def test_ppf(self, distribution, time, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).ppf(time, a0).shape == expected_out_shape(time=time, a0=a0)

    def test_ichf(self, distribution, probability, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).ichf(probability, a0).shape == expected_out_shape(
            probability=probability, a0=a0
        )

    def test_isf(self, distribution, probability, a0, expected_out_shape):
        out_shape = expected_out_shape(probability=probability, a0=a0)
        assert LeftTruncatedModel(distribution).isf(probability, a0).shape == out_shape
        assert LeftTruncatedModel(distribution).isf(np.full(out_shape, 0.5), a0) == approx(
            np.broadcast_to(LeftTruncatedModel(distribution).median(a0), out_shape)
        )

    def test_moment(self, distribution, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).moment(1, a0).shape == expected_out_shape(a0=a0)
        assert LeftTruncatedModel(distribution).moment(2, a0).shape == expected_out_shape(a0=a0)

    def test_mean(self, distribution, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).mean(a0).shape == expected_out_shape(a0=a0)

    def test_var(self, distribution, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).var(a0).shape == expected_out_shape(a0=a0)

    def test_median(self, distribution, a0, expected_out_shape):
        assert LeftTruncatedModel(distribution).median(a0).shape == expected_out_shape(a0=a0)

    def test_ls_integrate(self, distribution, integration_bound_a, integration_bound_b, a0, expected_out_shape):
        # integral_a^b dF(x)
        out_shape = expected_out_shape(
            integration_bound_a=integration_bound_a, integration_bound_b=integration_bound_b, a0=a0
        )
        integration = LeftTruncatedModel(distribution).ls_integrate(
            np.ones_like, integration_bound_a, integration_bound_b, a0, deg=100
        )
        assert integration.shape == out_shape
        assert integration == approx(
            LeftTruncatedModel(distribution).cdf(integration_bound_b, a0)
            - LeftTruncatedModel(distribution).cdf(integration_bound_a, a0)
        )
        # integral_0^inf x*dF(x)
        integration = LeftTruncatedModel(distribution).ls_integrate(
            lambda x: x, np.zeros_like(integration_bound_a), np.full_like(integration_bound_b, np.inf), a0, deg=100
        )
        assert integration == approx(np.broadcast_to(LeftTruncatedModel(distribution).mean(a0), out_shape), rel=1e-3)


class TestLeftTruncatedRegression:

    # def test_args_nb_assets(self, regression):
    #     frozen_distribution = regression.freeze()
    #     assert frozen_distribution.args_nb_assets == 1

    def test_rvs(self, regression, rvs_size, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).rvs(rvs_size, a0, covar).shape == expected_out_shape(
            a0=a0, covar=covar, size=rvs_size
        )
        assert all(
            arr.shape == expected_out_shape(a0=a0, covar=covar, size=rvs_size)
            for arr in LeftTruncatedModel(regression).rvs(rvs_size, a0, covar, return_event=True)
        )
        assert all(
            arr.shape == expected_out_shape(a0=a0, covar=covar, size=rvs_size)
            for arr in LeftTruncatedModel(regression).rvs(rvs_size, a0, covar, return_entry=True)
        )
        assert all(
            arr.shape == expected_out_shape(a0=a0, covar=covar, size=rvs_size)
            for arr in LeftTruncatedModel(regression).rvs(rvs_size, a0, covar, return_event=True, return_entry=True)
        )

    def test_sf(self, regression, time, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).sf(time, a0, covar).shape == expected_out_shape(
            time=time, a0=a0, covar=covar
        )

    def test_hf(self, regression, time, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).hf(time, a0, covar).shape == expected_out_shape(
            time=time, a0=a0, covar=covar
        )

    def test_chf(self, regression, time, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).chf(time, a0, covar).shape == expected_out_shape(
            time=time, a0=a0, covar=covar
        )

    def test_cdf(self, regression, time, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).cdf(time, a0, covar).shape == expected_out_shape(
            time=time, a0=a0, covar=covar
        )

    def test_pdf(self, regression, time, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).pdf(time, a0, covar).shape == expected_out_shape(
            time=time, a0=a0, covar=covar
        )

    def test_ppf(self, regression, time, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).ppf(time, a0, covar).shape == expected_out_shape(
            time=time, a0=a0, covar=covar
        )

    def test_ichf(self, regression, probability, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).ichf(probability, a0, covar).shape == expected_out_shape(
            probability=probability, a0=a0, covar=covar
        )

    def test_isf(self, regression, probability, a0, covar, expected_out_shape):
        out_shape = expected_out_shape(probability=probability, a0=a0, covar=covar)
        assert LeftTruncatedModel(regression).isf(probability, a0, covar).shape == out_shape
        assert LeftTruncatedModel(regression).isf(np.full(out_shape, 0.5), a0, covar) == approx(
            np.broadcast_to(LeftTruncatedModel(regression).median(a0, covar), out_shape)
        )

    def test_moment(self, regression, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).moment(1, a0, covar).shape == expected_out_shape(a0=a0, covar=covar)
        assert LeftTruncatedModel(regression).moment(2, a0, covar).shape == expected_out_shape(a0=a0, covar=covar)

    def test_mean(self, regression, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).mean(a0, covar).shape == expected_out_shape(a0=a0, covar=covar)

    def test_var(self, regression, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).var(a0, covar).shape == expected_out_shape(a0=a0, covar=covar)

    def test_median(self, regression, a0, covar, expected_out_shape):
        assert LeftTruncatedModel(regression).median(a0, covar).shape == expected_out_shape(a0=a0, covar=covar)

    def test_ls_integrate(self, regression, integration_bound_a, integration_bound_b, a0, covar, expected_out_shape):
        # integral_a^b dF(x)
        out_shape = expected_out_shape(
            integration_bound_a=integration_bound_a, integration_bound_b=integration_bound_b, a0=a0, covar=covar
        )
        np.random.seed(10)
        a0 = np.random.uniform(2.0, 8.0, size=a0.shape)
        integration = LeftTruncatedModel(regression).ls_integrate(
            np.ones_like, integration_bound_a, integration_bound_b, a0, covar, deg=100
        )
        assert integration.shape == out_shape
        assert integration == approx(
            LeftTruncatedModel(regression).cdf(integration_bound_b, a0, covar)
            - LeftTruncatedModel(regression).cdf(integration_bound_a, a0, covar)
        )
        # integral_0^inf x*dF(x)
        integration = LeftTruncatedModel(regression).ls_integrate(
            lambda x: x,
            np.zeros_like(integration_bound_a),
            np.full_like(integration_bound_b, np.inf),
            a0,
            covar,
            deg=100,
        )
        assert integration == approx(
            np.broadcast_to(LeftTruncatedModel(regression).mean(a0, covar), out_shape), rel=1e-3
        )



from abc import ABC, abstractmethod
from typing import Literal, Protocol

import numpy as np
from numpy.typing import NDArray


def cost(
    cf: float | NDArray[np.float64] = 0.0,
    cp: float | NDArray[np.float64] = 0.0,
    cr: float | NDArray[np.float64] = 0.0,
) -> NDArray[np.void]:
    struct_dtype = np.dtype(
        [
            ("cf", np.float64),
            ("cp", np.float64),
            ("cr", np.float64),
        ]
    )
    kwargs = {"cf": np.asarray(cf), "cp": np.asarray(cp), "cr": np.asarray(cr)}
    nb_assets = max(v.size if v.ndim > 0 else 0 for v in kwargs.values())
    shape: tuple[int, Literal[1]] | tuple[()] = (nb_assets, 1) if nb_assets > 0 else ()
    struct_cost = np.zeros(shape, dtype=struct_dtype)
    for k, v in kwargs.items():
        if v.ndim > 0:
            v = v.reshape(-1, 1)
        struct_cost[k] = v
    return struct_cost


class Reward(ABC):
    _cost_array: NDArray[np.void]

    @abstractmethod
    def conditional_expectation(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        """Conditional expected reward"""
        pass

    @abstractmethod
    def sample(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        """Reward conditional sampling.

        Parameters
        ----------
        time : ndarray
            Time duration values.


        Returns
        -------
        ndarray
            Random drawing of a reward with respect to time.
        """

    @property
    def nb_assets(self) -> int:
        """
        Number of assets.
        """
        if self._cost_array.shape == ():
            return 1
        return self._cost_array.shape[0]

    @property
    def ndim(self) -> int:
        """
        Cost number of dimension (either 0 or 1)
        """
        return self._cost_array.ndim

    @property
    def size(self) -> int:
        """
        Cost size
        """
        return self._cost_array.size


class RunToFailureReward(Reward):
    # noinspection PyUnresolvedReferences
    r"""Run-to-failure reward.

    Parameters
    ----------
    cf : float or 1darray
        The cost(s) of failure

    Attributes
    ----------
    cf
    """

    def __init__(self, cf: float | NDArray[np.float64]):
        self._cost_array = cost(cf=cf)

    @property
    def cf(self) -> NDArray[np.float64]:
        """
        Cost of failures

        Returns
        -------
        ndarray
        """
        return self._cost_array["cf"]

    @cf.setter
    def cf(self, value: float | NDArray[np.float64]) -> None:
        self._cost_array["cf"] = value

    def conditional_expectation(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.ones_like(time) * self.cf

    def sample(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        return self.conditional_expectation(time)


class AgeReplacementReward(Reward):
    # noinspection PyUnresolvedReferences
    r"""Age replacement reward.

    Parameters
    ----------
    cf : float or 1darray
        The cost(s) of failure
    cp : float or 1darray
        The cost(s) of preventive replacement

    Attributes
    ----------
    cf
    cp
    ar

    """

    def __init__(
        self, cf: float | NDArray[np.float64], cp: float | NDArray[np.float64], ar: float | NDArray[np.float64]
    ):
        self._cost_array = cost(cf=cf, cp=cp)
        ar = np.asarray(ar, dtype=np.float64)
        shape = () if ar.ndim == 0 else (ar.size, 1)
        self._ar = ar.reshape(shape)

    @property
    def cf(self) -> NDArray[np.float64]:
        """
        Cost of failures

        Returns
        -------
        ndarray
        """
        return self._cost_array["cf"]

    @property
    def cp(self) -> NDArray[np.float64]:
        """
        Cost of preventive replacement

        Returns
        -------
        ndarray
        """
        return self._cost_array["cp"]

    @cf.setter
    def cf(self, value: float | NDArray[np.float64]) -> None:
        self._cost_array["cf"] = value

    @cp.setter
    def cp(self, value: float | NDArray[np.float64]) -> None:
        self._cost_array["cp"] = value

    @property
    def ar(self) -> float | NDArray[np.float64]:
        """
        Age of replacement

        Returns
        -------
        ndarray
        """
        return self._ar

    @ar.setter
    def ar(self, value: float | NDArray[np.float64]) -> None:
        shape = () if value.ndim == 0 else (value.size, 1)
        self._ar = value.reshape(shape)

    def conditional_expectation(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.where(time < self.ar, self.cf, self.cp)

    def sample(self, time: NDArray[np.float64]) -> NDArray[np.float64]:
        return self.conditional_expectation(time)


class Discounting(Protocol):
    rate: float

    @abstractmethod
    def factor(self, timeline: NDArray[np.float64]) -> NDArray[np.float64]: ...

    @abstractmethod
    def annuity_factor(self, timeline: NDArray[np.float64]) -> NDArray[np.float64]: ...


class ExponentialDiscounting:
    rate: float

    def __init__(self, rate: float = 0.0):
        if rate < 0.0:
            raise ValueError(f"Invalid rate value. It must be positive. Got {rate}")
        self.rate = rate

    def factor(
        self,
        timeline: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        if self.rate != 0.0:
            return np.exp(-self.rate * timeline)
        else:
            return np.ones_like(timeline)

    def annuity_factor(
        self,
        timeline: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        if self.rate != 0.0:
            return (1 - np.exp(-self.rate * timeline)) / self.rate
        else:
            return timeline



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    overload,
)

import numpy as np
from numpy.typing import NDArray


@overload
def check_and_broadcast_bounds(
    a: float | NDArray[np.float64],
    b: None = None,
) -> NDArray[np.float64]: ...


@overload
def check_and_broadcast_bounds(
    a: float | NDArray[np.float64],
    b: float | NDArray[np.float64],
) -> tuple[NDArray[np.float64], NDArray[np.float64]]: ...


def check_and_broadcast_bounds(
    a: float | NDArray[np.float64],
    b: Optional[float | NDArray[np.float64]] = None,
) -> NDArray[np.float64] | tuple[NDArray[np.float64], NDArray[np.float64]]:

    def control_shape(bound: float | NDArray[np.float64]) -> NDArray[np.float64]:
        arr = np.asarray(bound, dtype=np.float64)
        if np.any(arr < 0):
            raise ValueError("Bound values of the integral can't be lower than 0")
        if arr.ndim > 2:
            raise ValueError("Bound the integral can't have more than 2 dimensions")
        return arr

    a = control_shape(a)
    if b is not None:
        b = control_shape(b)
        try:
            a, b = np.broadcast_arrays(a, b)
            return a.copy(), b.copy()
        except ValueError as err:
            raise ValueError(f"Incompatible a, b shapes. Got a.shape, b.shape : {a.shape}, {b.shape}") from err
    return a


def legendre_quadrature(
    func: Callable[[float | NDArray[np.float64]], np.float64 | NDArray[np.float64]],
    a: float | NDArray[np.float64],
    b: float | NDArray[np.float64],
    deg: int = 10,
) -> np.float64 | NDArray[np.float64]:
    r"""Numerical integration of `func` over the interval `[a,b]`

    `func` must accept (deg,), (deg, n) or (deg, m, n) array shapes
    a can be zero
    b must not be inf

    a, b shapes can be either 0d (float like), 1d or 2d
    """
    arr_a, arr_b = check_and_broadcast_bounds(a, b)  # () or (n,) or (m, n)
    quad = np.polynomial.legendre.leggauss(deg)  # (deg,)
    x = quad[0].reshape((-1,) + (1,) * arr_a.ndim)  # (deg,), (deg, 1) or (deg, 1, 1)
    w = quad[1].reshape((-1,) + (1,) * arr_a.ndim)  # (deg,), (deg, 1) or (deg, 1, 1)

    if np.any(arr_b == np.inf):
        raise ValueError("Bound values of Legendre quadrature must be finite")
    if np.any(arr_a > arr_b):
        raise ValueError("Bound values a must be lower than values of b")

    p = (arr_b - arr_a) / 2  # () or (n,) or (m, n)
    m = (arr_a + arr_b) / 2  # () or (n,) or (m, n)
    u = p * x + m  # (deg,) or (deg, n) or (deg, m, n)
    v = p * w  # (deg,) or (deg, n) or (deg, m, n)
    fvalues = func(u)  # (d_1, ..., d_i, deg) or (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
    if fvalues.shape[-len(u.shape) :] != u.shape:
        raise ValueError(
            f"""
            func can't squeeze input dimensions. If x has shape (d_1, ..., d_i), func(x) must have shape (..., d_1, ..., d_i).
            Ex : if x.shape == (m, n), func(x).shape == (..., m, n).
            """
        )

    return np.sum(v * fvalues, axis=-v.ndim)  # (d_1, ..., d_i) or (d_1, ..., d_i, n) or (d_1, ..., d_i, m, n)


def laguerre_quadrature(
    func: Callable[[float | NDArray[np.float64]], np.float64 | NDArray[np.float64]],
    a: float | NDArray[np.float64] = 0.0,
    deg: int = 10,
) -> np.float64 | NDArray[np.float64]:
    r"""Numerical integration of `func * exp(-x)` over the interval `[a, inf]`

    `func` must accept (deg,), (deg, n) or (deg, m, n) array shapes
    It must handle at least 3 dimensions.
    a can be zero with ndim <= 2.
    """
    arr_a = check_and_broadcast_bounds(a)  # () or (n,) or (m, n)
    quad = np.polynomial.laguerre.laggauss(deg)  # (deg,)
    x = quad[0].reshape((-1,) + (1,) * arr_a.ndim)  # (deg,), (deg, 1) or (deg, 1, 1)
    w = quad[1].reshape((-1,) + (1,) * arr_a.ndim)  # (deg,), (deg, 1) or (deg, 1, 1)

    shifted_x = x + arr_a  # (deg,) or (deg, n) or (deg, m, n)
    fvalues = func(shifted_x)  # (d_1, ..., d_i, deg) or (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
    if fvalues.shape[-len(shifted_x.shape) :] != shifted_x.shape:
        # func est une fonction réel univariée et pas multivariée
        raise ValueError(
            f"""
            func can't squeeze input dimensions. If x has shape (d_1, ..., d_i), func(x) must have shape (..., d_1, ..., d_i).
            Ex : if x.shape == (m, n), func(x).shape == (..., m, n).
            """
        )

    exp_a = np.where(np.exp(-arr_a) == 0, 1.0, np.exp(-arr_a))  # () or (n,) or (m, n)
    return np.sum(
        w * fvalues * exp_a, axis=-shifted_x.ndim
    )  # (d_1, ..., d_i) or (d_1, ..., d_i, n) or (d_1, ..., d_i, m, n)


def unweighted_laguerre_quadrature(
    func: Callable[[float | NDArray[np.float64]], np.float64 | NDArray[np.float64]],
    a: float | NDArray[np.float64] = 0.0,
    deg: int = 10,
) -> np.float64 | NDArray[np.float64]:
    r"""Numerical integration of `func` over the interval `[a, inf]`

    `func` must accept (deg,), (deg, n) or (deg, m, n) array shapes
    It must handle at least 3 dimensions.
    a can be zero with ndim <= 2.
    """

    quad = np.polynomial.laguerre.laggauss(deg)  # (deg,)
    arr_a = check_and_broadcast_bounds(a)  # () or (n,) or (m, n)
    x = quad[0].reshape((-1,) + (1,) * arr_a.ndim)  # (deg,), (deg, 1) or (deg, 1, 1)
    w = quad[1].reshape((-1,) + (1,) * arr_a.ndim)  # (deg,), (deg, 1) or (deg, 1, 1)

    shifted_x = x + arr_a  # (deg,) or (deg, n) or (deg, m, n)
    fvalues = func(shifted_x)  # (d_1, ..., d_i, deg) or (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
    if fvalues.shape[-len(shifted_x.shape) :] != shifted_x.shape:
        raise ValueError(
            f"""
            func can't squeeze input dimensions. If x has shape (d_1, ..., d_i), func(x) must have shape (..., d_1, ..., d_i).
            Ex : if x.shape == (m, n), func(x).shape == (..., m, n).
            """
        )
    return np.sum(
        w * fvalues * np.exp(x), axis=-shifted_x.ndim
    )  # (d_1, ..., d_i) or (d_1, ..., d_i, n) or (d_1, ..., d_i, m, n)



import copy

import numpy as np
from numpy.typing import NDArray
from typing_extensions import override

from relife.economic import (
    ExponentialDiscounting,
    Reward,
)
from relife.stochastic_process.renewal_equations import (
    delayed_renewal_equation_solver,
    renewal_equation_solver,
)

from ._sample import RenewalProcessSample, RenewalRewardProcessSample
from .base import StochasticProcess


class RenewalProcess(StochasticProcess):
    # noinspection PyUnresolvedReferences
    """Renewal process.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.

    first_lifetime_model : any lifetime distribution or frozen lifetime model, optional
        A lifetime model for the first renewal (delayed renewal process). It is None by default

    Attributes
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.

    first_lifetime_model : any lifetime distribution or frozen lifetime model, optional
        A lifetime model for the first renewal (delayed renewal process). It is None by default
    nb_params
    params
    params_names
    """

    def __init__(self, lifetime_model, first_lifetime_model=None):
        super().__init__()
        self.lifetime_model = lifetime_model
        self.first_lifetime_model = first_lifetime_model

    def _make_timeline(self, tf, nb_steps):
        # tile is necessary to ensure broadcasting of the operations
        timeline = np.linspace(0, tf, nb_steps, dtype=np.float64)  # (nb_steps,)
        args_nb_assets = getattr(self.lifetime_model, "nb_assets", 1)
        if args_nb_assets > 1:
            timeline = np.tile(timeline, (args_nb_assets, 1))
        return timeline  # (nb_steps,) or (m, nb_steps)

    def renewal_function(self, tf, nb_steps):
        r"""The renewal function.

        Parameters
        ----------
        tf : float
            Time horizon. The renewal function will be computed up until this calendar time.
        nb_steps : int
            The number of steps used to compute the renewal function.

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline used to compute the renewal function and its corresponding values at each
            step of the timeline.

        Notes
        -----
        The expected total number of renewals is computed  by solving the
        renewal equation:

        .. math::

            m(t) = F_1(t) + \int_0^t m(t-x) \mathrm{d}F(x)

        where:

        - :math:`m` is the renewal function,
        - :math:`F` is the cumulative distribution function of the underlying
          lifetime model,
        - :math:`F_1` is the cumulative distribution function of the underlying
          lifetime model for the fist renewal in the case of a delayed renewal
          process.

        References
        ----------
        .. [1] Rausand, M., Barros, A., & Hoyland, A. (2020). System Reliability
            Theory: Models, Statistical Methods, and Applications. John Wiley &
            Sons.
        """

        timeline = self._make_timeline(tf, nb_steps)  # (nb_steps,) or (1, nb_steps)
        renewal_function = renewal_equation_solver(
            timeline,
            self.lifetime_model,
            self.lifetime_model.cdf if self.first_lifetime_model is None else self.first_lifetime_model.cdf,
        )
        if timeline.ndim == 2:
            return timeline[0, :], renewal_function
        return timeline, renewal_function

    def renewal_density(self, tf, nb_steps):
        r"""The renewal density.

        Parameters
        ----------
        tf : float
            Time horizon. The renewal density will be computed up until this calendar time.
        nb_steps : int
            The number of steps used to compute the renewal density.

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline used to compute the renewal density and its corresponding values at each
            step of the timeline.

        Notes
        -----
        The renewal density is the derivative of the renewal function with
        respect to time. It is computed by solving the renewal equation:

        .. math::

            \mu(t) = f_1(t) + \int_0^t \mu(t-x) \mathrm{d}F(x)

        where:

        - :math:`\mu` is the renewal function,
        - :math:`F` is the cumulative distribution function of the underlying
          lifetime model,
        - :math:`f_1` is the probability density function of the underlying
          lifetime model for the fist renewal in the case of a delayed renewal
          process.

        References
        ----------
        .. [1] Rausand, M., Barros, A., & Hoyland, A. (2020). System Reliability
            Theory: Models, Statistical Methods, and Applications. John Wiley &
            Sons.
        """
        timeline = self._make_timeline(tf, nb_steps)  #  (nb_steps,) or (m, nb_steps)
        renewal_density = renewal_equation_solver(
            timeline,
            self.lifetime_model,
            self.lifetime_model.pdf if self.first_lifetime_model is None else self.first_lifetime_model.pdf,
        )
        if timeline.ndim == 2:
            return timeline[0, :], renewal_density
        return timeline, renewal_density

    def sample(self, size, tf, t0=0.0, seed=None):
        """Renewal data sampling.

        This function will sample data and encapsulate them in an object.

        Parameters
        ----------
        size : int
            The size of the desired sample
        tf : float
            Time at the end of the observation.
        t0 : float, default 0
            Time at the beginning of the observation.
        size : int or tuple of 2 int
            Size of the sample
        seed : int, optional
            Random seed, by default None.

        """

        from ._sample import RenewalProcessIterable

        iterable = RenewalProcessIterable(self, size, tf, t0=t0, seed=seed)
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(struct_array, order=("sample_id", "asset_id", "timeline"))
        return RenewalProcessSample(t0, tf, struct_array)

    def generate_failure_data(self, size, tf, t0=0.0, seed=None):
        """Generate lifetime data

        This function will generate lifetime data that can be used to fit a lifetime model.

        Parameters
        ----------
        size : int
            The size of the desired sample
        tf : float
            Time at the end of the observation.
        t0 : float, default 0
            Time at the beginning of the observation.
        seed : int, optional
            Random seed, by default None.

        Returns
        -------
        A dict of time, event, entry and args (covariates)

        """
        from ._sample import RenewalProcessIterable

        if self.first_lifetime_model is not None and self.first_lifetime_model != self.lifetime_model:
            from relife.lifetime_model import FrozenLeftTruncatedModel

            if (
                isinstance(self.first_lifetime_model, FrozenLeftTruncatedModel)
                and self.first_lifetime_model.unfreeze() == self.lifetime_model
            ):
                pass
            else:
                raise ValueError(
                    f"Calling sample_lifetime_data with lifetime_model different from first_lifetime_model is ambiguous."
                )
        iterable = RenewalProcessIterable(self, size, tf, t0=t0, seed=seed)
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(struct_array, order=("sample_id", "asset_id", "timeline"))

        nb_assets = int(np.max(struct_array["asset_id"])) + 1
        args_2d = tuple((np.atleast_2d(arg) for arg in getattr(self.lifetime_model, "args", ())))
        broadcasted_args = tuple((np.broadcast_to(arg, (nb_assets, arg.shape[-1])) for arg in args_2d))
        tuple_args_arr = tuple((np.take(np.asarray(arg), struct_array["asset_id"]) for arg in broadcasted_args))

        returned_dict = {
            "time": struct_array["time"].copy(),
            "event": struct_array["event"].copy(),
            "entry": struct_array["entry"].copy(),
            "args": tuple_args_arr,
        }
        return returned_dict


class RenewalRewardProcess(RenewalProcess):
    # noinspection PyUnresolvedReferences
    """Renewal reward process.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    reward : Reward
        A reward object that answers costs or conditional costs given lifetime values
    discounting_rate : float
        The discounting rate value used in the exponential discounting function
    first_lifetime_model : any lifetime distribution or frozen lifetime model, optional
        A lifetime model for the first renewal (delayed renewal process). It is None by default
    reward : Reward
        A reward object for the first renewal

    Attributes
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    first_lifetime_model : any lifetime distribution or frozen lifetime model, optional
        A lifetime model for the first renewal (delayed renewal process). It is None by default
    reward : Reward
        A reward object that answers costs or conditional costs given lifetime values
    first_reward : Reward
        A reward object for the first renewal. If it is not given at the initialization, it is a copy of reward.
    discounting_rate
    nb_params
    params
    params_names
    """

    def __init__(self, lifetime_model, reward, discounting_rate=0.0, first_lifetime_model=None, first_reward=None):
        super().__init__(lifetime_model, first_lifetime_model)
        self.reward = reward
        self.first_reward = first_reward if first_reward is not None else copy.deepcopy(reward)
        self.discounting = ExponentialDiscounting(discounting_rate)

    @property
    def discounting_rate(self):
        """
        The discounting rate value
        """
        return self.discounting.rate

    @discounting_rate.setter
    def discounting_rate(self, value):
        """
        The discounting rate value setter

        Parameters
        ----------
        value : float
            The new discounting rate value to be set
        """
        self.discounting.rate = value

    @override
    def _make_timeline(self, tf, nb_steps):
        # tile is necessary to ensure broadcasting of the operations
        timeline = np.linspace(0, tf, nb_steps, dtype=np.float64)  # (nb_steps,)
        args_nb_assets = getattr(self.lifetime_model, "nb_assets", 1)  # default 1 for LifetimeDistribution case
        if args_nb_assets > 1:
            timeline = np.tile(timeline, (args_nb_assets, 1))
        elif self.reward.ndim == 2:  # elif because we consider that if m > 1 in frozen_model, in reward it is 1 or m
            timeline = np.tile(timeline, (self.reward.size, 1))
        return timeline  # (nb_steps,) or (m, nb_steps)

    def expected_total_reward(self, tf, nb_steps):
        r"""The expected total reward.

        The renewal equation solved to compute the expected reward is:

        .. math::

            z(t) = \int_0^t E[Y | X = x] e^{-\delta x} \mathrm{d}F(x) + \int_0^t z(t-x)
            e^{-\delta x}\mathrm{d}F(x)

        where:

        - :math:`z` is the expected total reward,
        - :math:`F` is the cumulative distribution function of the underlying
          lifetime model,
        - :math:`X` the interarrival random variable,
        - :math:`Y` the associated reward,
        - :math:`D` the exponential discount factor.

        If the renewal reward process is delayed, the expected total reward is
        modified as:

        .. math::

            z_1(t) = \int_0^t E[Y_1 | X_1 = x] e^{-\delta x} \mathrm{d}F_1(x) + \int_0^t
            z(t-x) e^{-\delta x} \mathrm{d}F_1(x)

        where:

        - :math:`z_1` is the expected total reward with delay,
        - :math:`F_1` is the cumulative distribution function of the lifetime
          model for the first renewal,
        - :math:`X_1` the interarrival random variable of the first renewal,
        - :math:`Y_1` the associated reward of the first renewal,

        Parameters
        ----------
        tf : float
            Time horizon. The expected total reward will be computed up until this calendar time.
        nb_steps : int
            The number of steps used to compute the expected total reward.

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline used to compute the expected total reward and its corresponding values at each
            step of the timeline.
        """
        timeline = self._make_timeline(tf, nb_steps)  #  (nb_steps,) or (m, nb_steps)
        z = renewal_equation_solver(
            timeline,
            self.lifetime_model,
            lambda t: self.lifetime_model.ls_integrate(
                lambda x: self.reward.conditional_expectation(x) * self.discounting.factor(x),
                np.zeros_like(t),
                t,
                deg=15,
            ),  # reward partial expectation
            discounting=self.discounting,
        )
        if self.first_lifetime_model is not None:
            z = delayed_renewal_equation_solver(
                timeline,
                z,
                self.first_lifetime_model,
                lambda t: self.first_lifetime_model.ls_integrate(
                    lambda x: self.first_reward.conditional_expectation(x) * self.discounting.factor(x),
                    np.zeros_like(t),
                    t,
                    deg=15,
                ),  # reward partial expectation
                discounting=self.discounting,
            )
        if timeline.ndim == 2:
            return timeline[0, :], z  # (nb_steps,) and (m, nb_steps)
        return timeline, z  # (nb_steps,) and (nb_steps, )

    def asymptotic_expected_total_reward(self):
        r"""Asymptotic expected total reward.

        The asymptotic expected total reward is:

        .. math::

            z^\infty = \lim_{t\to \infty} z(t) = \dfrac{E\left[Y e^{-\delta X}\right]}{1-E\left[e^{-\delta X}\right]}

        where:

        - :math:`X` the interarrival random variable,
        - :math:`Y` the associated reward,
        - :math:`D` the exponential discount factor.

        If the renewal reward process is delayed, the asymptotic expected total
        reward is modified as:

        .. math::

            z_1^\infty = E\left[Y_1 e^{-\delta X_1}\right] + z^\infty E\left[e^{-\delta X_1}\right]

        where:

        - :math:`X_1` the interarrival random variable of the first renewal,
        - :math:`Y_1` the associated reward of the first renewal,

        Returns
        -------
        ndarray
            The assymptotic expected total reward of the process.
        """
        lf = self.lifetime_model.ls_integrate(
            lambda x: self.discounting.factor(x), 0.0, np.inf, deg=100
        )  # () or (m, 1)
        if self.discounting_rate == 0.0:
            return np.full_like(np.squeeze(lf), np.inf)
        ly = self.lifetime_model.ls_integrate(
            lambda x: self.discounting.factor(x) * self.reward.conditional_expectation(x), 0.0, np.inf, deg=100
        )
        z = ly / (1 - lf)  # () or (m, 1)
        if self.first_lifetime_model is not None:
            lf1 = self.first_lifetime_model.ls_integrate(lambda x: self.discounting.factor(x), 0.0, np.inf, deg=100)
            ly1 = self.first_lifetime_model.ls_integrate(
                lambda x: self.discounting.factor(x) * self.first_reward.conditional_expectation(x),
                0.0,
                np.inf,
                deg=100,
            )
            z = ly1 + z * lf1  # () or (m, 1)
        return np.squeeze(z)  # () or (m,)

    def expected_equivalent_annual_worth(self, tf, nb_steps):
        """Expected equivalent annual worth.

        Gives the equivalent annual worth of the expected total reward of the
        process at each point of the timeline.

        The equivalent annual worth at time :math:`t` is equal to the expected
        total reward :math:`z` divided by the annuity factor :math:`AF(t)`.

        Parameters
        ----------
        tf : float
            Time horizon. The expected equivalent annual worth will be computed up until this calendar time.
        nb_steps : int
            The number of steps used to compute the expected equivalent annual worth.

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline used to compute the expected equivalent annual worth and its corresponding values at each
            step of the timeline.
        """
        timeline, z = self.expected_total_reward(
            tf, nb_steps
        )  # timeline : (nb_steps,) z : (nb_steps,) or (m, nb_steps)
        af = self.discounting.annuity_factor(timeline)  # (nb_steps,)
        if z.ndim == 2 and af.shape != z.shape:  # (m, nb_steps)
            af = np.tile(af, (z.shape[0], 1))  # (m, nb_steps)
        q = z / (af + 1e-6)  # # (nb_steps,) or (m, nb_steps) avoid zero division
        if self.first_lifetime_model is not None:
            q0 = self.first_reward.conditional_expectation(0.0) * self.first_lifetime_model.pdf(0.0)
        else:
            q0 = self.reward.conditional_expectation(0.0) * self.lifetime_model.pdf(0.0)
        # q0 : () or (m, 1)
        q0 = np.broadcast_to(q0, af.shape)  # (), (nb_steps,) or (m, nb_steps)
        eeac = np.where(af == 0, q0, q)  # (nb_steps,) or (m, nb_steps)
        if timeline.ndim == 2:
            return timeline[0, :], eeac  # (nb_steps,) and (m, nb_steps)
        return timeline, eeac  # (nb_steps,) and (nb_steps)

    def asymptotic_expected_equivalent_annual_worth(self) -> NDArray[np.float64]:
        """Asymptotic expected equivalent annual worth.

        Returns
        -------
        ndarray
            The assymptotic expected equivalent annual worth
        """
        if self.discounting_rate == 0.0:
            return np.squeeze(
                self.lifetime_model.ls_integrate(lambda x: self.reward.conditional_expectation(x), 0.0, np.inf, deg=100)
                / self.lifetime_model.mean()
            )  # () or (m,)
        return np.squeeze(self.discounting_rate * self.asymptotic_expected_total_reward())  # () or (m,)

    @override
    def sample(self, size, tf, t0=0.0, seed=None):
        from ._sample import RenewalProcessIterable

        iterable = RenewalProcessIterable(self, size, tf, t0=t0, seed=seed)
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(struct_array, order=("nb_renewal", "asset_id", "sample_id"))
        return RenewalRewardProcessSample(t0, tf, struct_array)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Optional, TypeVar, Union

import numpy as np
from numpy.typing import NDArray
from scipy import stats
from scipy.optimize import approx_fprime

from relife import ParametricModel


class Likelihood(ABC):
    model: ParametricModel
    data: Any

    @abstractmethod
    def negative_log(self, params: NDArray[np.float64]) -> np.float64:
        """
        Negative log likelihood.

        Parameters
        ----------
        params : ndarray
            Parameters values on which likelihood is evaluated

        Returns
        -------
        float
            Negative log likelihood value
        """

    @abstractmethod
    def maximum_likelihood_estimation(self, **kwargs: Any) -> FittingResults: ...


L = TypeVar("L", bound="LikelihoodFromLifetimes")  # maybe other likelihood in the future


def hessian_cs(
    likelihood: L,
    params: NDArray[np.float64],
    eps: float = 1e-6,
) -> NDArray[np.float64]:

    size = params.size
    hess = np.empty((size, size))
    u = eps * 1j * np.eye(size)
    complex_params = params.astype(np.complex64)  # change params to complex
    for i in range(size):
        for j in range(i, size):
            hess[i, j] = np.imag(likelihood.jac_negative_log(complex_params + u[i])[j]) / eps
            if i != j:
                hess[j, i] = hess[i, j]
    return hess


def hessian_2point(
    likelihood: L,
    params: NDArray[np.float64],
    eps: float = 1e-6,
) -> NDArray[np.float64]:
    size = params.size
    hess = np.empty((size, size))
    for i in range(size):
        hess[i] = approx_fprime(
            params,
            lambda x: likelihood.jac_negative_log(x)[i],
            eps,
        )
    return hess


M = TypeVar("M", bound=Union["LifetimeDistribution", "LifetimeRegression", "MinimumDistribution"])


def approx_hessian(likelihood: L, params: NDArray[np.float64], eps: float = 1e-6) -> NDArray[np.float64]:

    def hessian_scheme(model: M):
        from relife.lifetime_model import Gamma
        from relife.lifetime_model.regression import LifetimeRegression

        if isinstance(model, LifetimeRegression):
            return hessian_scheme(model.baseline)
        if isinstance(model, Gamma):
            return hessian_2point
        return hessian_cs

    return hessian_scheme(likelihood.model)(likelihood, params, eps=eps)


@dataclass
class FittingResults:
    """Fitting results of the parametric_model core."""

    nb_obversations: int  #: Number of observations (samples)
    optimal_params: NDArray[np.float64] = field(repr=False)  #: Optimal parameters values
    neg_log_likelihood: np.float64 = field(repr=False)  #: Negative log likelihood value at optimal parameters values

    covariance_matrix: Optional[NDArray[np.floating[Any]]] = field(
        repr=False, default=None
    )  #: Covariance matrix (computed as the inverse of the Hessian matrix).

    nb_params: int = field(init=False, repr=False)  #: Number of parameters.
    AIC: float = field(init=False)  #: Akaike Information Criterion.
    AICc: float = field(init=False)  #: Akaike Information Criterion with a correction for small sample sizes.
    BIC: float = field(init=False)  #: Bayesian Information Criterion.
    se: Optional[NDArray[np.float64]] = field(
        init=False, repr=False
    )  #: Standard error, square root of the diagonal of the covariance matrix
    IC: Optional[NDArray[np.float64]] = field(init=False, repr=False)  #: 95% IC

    def __post_init__(self):
        nb_params = self.optimal_params.size
        self.AIC = float(2 * nb_params + 2 * self.neg_log_likelihood)
        self.AICc = float(self.AIC + 2 * nb_params * (nb_params + 1) / (self.nb_obversations - nb_params - 1))
        self.BIC = float(np.log(self.nb_obversations) * nb_params + 2 * self.neg_log_likelihood)

        self.se = None
        if self.covariance_matrix is not None:
            self.se = np.sqrt(np.diag(self.covariance_matrix))
            self.IC = self.optimal_params.reshape(-1, 1) + stats.norm.ppf((0.05, 0.95)) * self.se.reshape(
                -1, 1
            ) / np.sqrt(
                self.nb_obversations
            )  # (p, 2)

    def se_estimation_function(self, jac_f: np.ndarray) -> np.float64 | NDArray[np.float64]:
        """Standard error estimation function.

        Parameters
        ----------
        jac_f : 1D array
            The Jacobian of a function f with respect to params.

        Returns
        -------
        1D array
            Standard error for f(params).

        References
        ----------
        .. [1] Meeker, W. Q., Escobar, L. A., & Pascual, F. G. (2022).
            Statistical methods for reliability data. John Wiley & Sons.
        """
        # [1] equation B.10 in Appendix
        # jac_f : (p,), (p, n) or (p, m, n)
        # self.var : (p, p)
        if self.covariance_matrix is not None:
            return np.sqrt(np.einsum("p...,pp,p...", jac_f, self.covariance_matrix, jac_f))  # (), (n,) or (m, n)
        raise ValueError("Can't compute if var is None")

    def __str__(self) -> str:
        """Returns a string representation of FittingResults with fields in a single column."""
        fields = [("fitted params", self.optimal_params), ("AIC", self.AIC), ("AICc", self.AICc), ("BIC", self.BIC)]
        # Find the maximum field name length for alignment
        max_name_length = max(len(name) for name, _ in fields)
        lines = []
        for name, value in fields:
            # Format arrays to be more compact
            if isinstance(value, np.ndarray):
                value_str = f"[{', '.join(f'{x:.6g}' for x in value)}]"
            else:
                value_str = f"{value:.6g}" if isinstance(value, float) else str(value)
            lines.append(f"{name:<{max_name_length}} : {value_str}")
        return "\n".join(lines)



from relife.economic import AgeReplacementReward, RunToFailureReward, cost


def test_cost(cf, cp, expected_out_shape):
    cost_arr = cost(cf=cf, cp=cp)
    assert cost_arr["cf"].shape == expected_out_shape(cf=cf, cp=cp)


def test_run_to_failure_reward(time, cf, expected_out_shape):
    reward = RunToFailureReward(cf)
    assert reward.conditional_expectation(time).shape == expected_out_shape(time=time, cf=cf)
    assert reward.sample(time).shape == expected_out_shape(time=time, cf=cf)


def test_age_replacement_reward(time, cf, cp, ar, expected_out_shape):
    reward = AgeReplacementReward(cf, cp, ar)
    assert reward.conditional_expectation(time).shape == expected_out_shape(time=time, cf=cf, cp=cp, ar=ar)
    assert reward.sample(time).shape == expected_out_shape(time=time, cf=cf, cp=cp, ar=ar)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import minimize

from relife.data import LifetimeData

from . import FittingResults, approx_hessian
from ._base import Likelihood

if TYPE_CHECKING:
    from relife.lifetime_model import (
        MinimumDistribution,
    )

    from ..lifetime_model.distribution import LifetimeDistribution
    from ..lifetime_model.regression import LifetimeRegression


class LikelihoodFromLifetimes(Likelihood):
    """
    Generic likelihood object for parametric_model lifetime model

    Parameters
    ----------
    model : ParametricLifetimeDistribution
        Underlying core used to compute probability functions
    lifetime_data : LifetimeData
        Observed lifetime data used one which the likelihood is evaluated
    """

    def __init__(
        self,
        model: Union[LifetimeDistribution, LifetimeRegression, MinimumDistribution],
        lifetime_data: LifetimeData,
    ):
        self.model = model
        self.data = lifetime_data

    def _complete_contribs(self, lifetime_data: LifetimeData) -> Optional[np.float64]:
        if lifetime_data.complete is None:
            return None
        return -np.sum(
            np.log(
                self.model.hf(
                    lifetime_data.complete.lifetime_values,
                    *lifetime_data.complete.args,
                )
            )  # (m, 1)
        )  # ()

    def _right_censored_contribs(self, lifetime_data: LifetimeData) -> Optional[np.float64]:
        if lifetime_data.complete_or_right_censored is None:
            return None
        return np.sum(
            self.model.chf(
                lifetime_data.complete_or_right_censored.lifetime_values,
                *lifetime_data.complete_or_right_censored.args,
            ),
            dtype=np.float64,  # (m, 1)
        )  # ()

    def _left_censored_contribs(self, lifetime_data: LifetimeData) -> Optional[np.float64]:
        if lifetime_data.left_censoring is None:
            return None
        return -np.sum(
            np.log(
                -np.expm1(
                    -self.model.chf(
                        lifetime_data.left_censoring.lifetime_values,
                        *lifetime_data.left_censoring.args,
                    )
                )
            )  # (m, 1)
        )  # ()

    def _left_truncations_contribs(self, lifetime_data: LifetimeData) -> Optional[np.float64]:
        if lifetime_data.left_truncation is None:
            return None
        return -np.sum(
            self.model.chf(
                lifetime_data.left_truncation.lifetime_values,
                *lifetime_data.left_truncation.args,
            ),  # (m, 1)
            dtype=np.float64,
        )  # ()

    def _jac_complete_contribs(self, lifetime_data: LifetimeData) -> Optional[NDArray[np.float64]]:
        if lifetime_data.complete is None:
            return None
        return -np.sum(
            self.model.jac_hf(
                lifetime_data.complete.lifetime_values,
                *lifetime_data.complete.args,
                asarray=True,
            )  # (p, m, 1)
            / self.model.hf(
                lifetime_data.complete.lifetime_values,
                *lifetime_data.complete.args,
            ),  # (m, 1)
            axis=(1, 2),
        )  # (p,)

    def _jac_right_censored_contribs(self, lifetime_data: LifetimeData) -> Optional[NDArray[np.float64]]:
        if lifetime_data.complete_or_right_censored is None:
            return None
        return np.sum(
            self.model.jac_chf(
                lifetime_data.complete_or_right_censored.lifetime_values,
                *lifetime_data.complete_or_right_censored.args,
                asarray=True,
            ),  # (p, m, 1)
            axis=(1, 2),
        )  # (p,)

    def _jac_left_censored_contribs(self, lifetime_data: LifetimeData) -> Optional[NDArray[np.float64]]:
        if lifetime_data.left_censoring is None:
            return None
        return -np.sum(
            self.model.jac_chf(
                lifetime_data.left_censoring.lifetime_values,
                *lifetime_data.left_censoring.args,
                asarray=True,
            )  # (p, m, 1)
            / np.expm1(
                self.model.chf(
                    lifetime_data.left_censoring.lifetime_values,
                    *lifetime_data.left_censoring.args,
                )
            ),  # (m, 1)
            axis=(1, 2),
        )  # (p,)

    def _jac_left_truncations_contribs(self, lifetime_data: LifetimeData) -> Optional[NDArray[np.float64]]:
        if lifetime_data.left_truncation is None:
            return None
        return -np.sum(
            self.model.jac_chf(
                lifetime_data.left_truncation.lifetime_values,
                *lifetime_data.left_truncation.args,
                asarray=True,
            ),  # (p, m, 1)
            axis=(1, 2),
        )  # (p,)

    def negative_log(
        self,
        params: NDArray[np.float64],  # (p,)
    ) -> np.float64:
        model_params = np.copy(self.model.params)
        self.model.params = params  # changes model params
        contributions = (
            self._complete_contribs(self.data),
            self._right_censored_contribs(self.data),
            self._left_censored_contribs(self.data),
            self._left_truncations_contribs(self.data),
        )
        self.model.params = model_params  # reset model params (negative_log must not change model params)
        return sum(x for x in contributions if x is not None)  # ()

    def jac_negative_log(
        self,
        params: NDArray[np.float64],  # (p,)
    ) -> NDArray[np.float64]:
        """
        Jacobian of the negative log likelihood.

        The jacobian (here gradient) is computed with respect to parameters

        Parameters
        ----------
        params : ndarray
            Parameters values on which the jacobian is evaluated

        Returns
        -------
        ndarray
            Jacobian of the negative log likelihood value
        """
        model_params = np.copy(self.model.params)
        self.model.params = params  # changes model params
        jac_contributions = (
            self._jac_complete_contribs(self.data),
            self._jac_right_censored_contribs(self.data),
            self._jac_left_censored_contribs(self.data),
            self._jac_left_truncations_contribs(self.data),
        )
        self.model.params = model_params  # reset model params (jac_negative_log must not change model params)
        return sum(x for x in jac_contributions if x is not None)  # (p,)

    def maximum_likelihood_estimation(self, **options: Any) -> FittingResults:
        # configure and run the optimizer
        minimize_kwargs = {
            "method": options.get("method", "L-BFGS-B"),
            "constraints": options.get("constraints", ()),
            "bounds": options.get("bounds", None),
            "x0": options.get("x0", self.model.params),
        }
        optimizer = minimize(
            self.negative_log,
            minimize_kwargs.pop("x0"),
            jac=(
                self.jac_negative_log
                if minimize_kwargs["method"] not in ("Nelder-Mead", "Powell", "COBYLA", "COBYQA")
                else None
            ),
            **minimize_kwargs,
        )
        optimal_params = np.copy(optimizer.x)
        neg_log_likelihood = np.copy(optimizer.fun)  # neg_log_likelihood value at optimal
        hessian = approx_hessian(self, optimal_params)
        covariance_matrix = np.linalg.inv(hessian)
        return FittingResults(
            self.data.nb_samples, optimal_params, neg_log_likelihood, covariance_matrix=covariance_matrix
        )


# M = TypeVar("M", bound=Union["LifetimeDistribution", "LifetimeRegression", "MinimumDistribution"])
#
# # TODO : move into model interface as _init_params(self, *args)
# def _init_params_values(model: M, lifetime_data: LifetimeData) -> NDArray[np.float64]:
#     from relife.lifetime_model import Gompertz, LifetimeRegression
#
#     if isinstance(model, LifetimeRegression):
#         model.baseline.params = _init_params_values(model.baseline, lifetime_data)
#         param0 = np.zeros_like(model.params, dtype=np.float64)
#         param0[-model.baseline.params.size :] = model.baseline.params
#         return param0
#     elif isinstance(model, Gompertz):
#         param0 = np.empty(model.nb_params, dtype=np.float64)
#         rate = np.pi / (np.sqrt(6) * np.std(lifetime_data.complete_or_right_censored.lifetime_values))
#         shape = np.exp(-rate * np.mean(lifetime_data.complete_or_right_censored.lifetime_values))
#         param0[0] = shape
#         param0[1] = rate
#         return param0
#     else:  # other cases : Weibull, Gamma, Exponential, ...
#         if lifetime_data.complete_or_right_censored is not None:
#             param0 = np.ones(model.nb_params, dtype=np.float64)
#             param0[-1] = 1 / np.median(lifetime_data.complete_or_right_censored.lifetime_values)
#             return param0
#         return np.zeros(model.nb_params, dtype=np.float64)
#
# # TODO : move into model interface as _params_bounds(self, *args)
# def _get_params_bounds(model: M) -> Bounds:
#     from relife.lifetime_model import LifetimeRegression
#
#     if isinstance(model, LifetimeRegression):
#         lb = np.concatenate(
#             (
#                 np.full(model.covar_effect.nb_params, -np.inf),
#                 _get_params_bounds(model.baseline).lb,
#             )
#         )
#         ub = np.concatenate(
#             (
#                 np.full(model.covar_effect.nb_params, np.inf),
#                 _get_params_bounds(model.baseline).ub,
#             )
#         )
#         return Bounds(lb, ub)
#     else:
#         return Bounds(
#             np.full(model.nb_params, np.finfo(float).resolution),
#             np.full(model.nb_params, np.inf),
#         )



from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    Optional,
    Self,
    overload,
)

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import Bounds

from relife import FrozenParametricModel as FrozenParametricModel
from relife import ParametricModel as ParametricModel
from relife._typing import _B, _X, _Y, _Xs
from relife.likelihood import FittingResults as FittingResults

from ._plot import PlotParametricLifetimeModel

class ParametricLifetimeModel(ParametricModel, ABC, Generic[*_Xs]):
    @overload
    def sf(self, time: _X, *args: *_Xs) -> _Y: ...
    @abstractmethod
    def hf(self, time: _X, *args: *_Xs) -> _Y: ...
    @abstractmethod
    def chf(self, time: _X, *args: *_Xs) -> _Y: ...
    @abstractmethod
    def pdf(self, time: _X, *args: *_Xs) -> _Y: ...
    def cdf(self, time: _X, *args: *_Xs) -> _Y: ...
    def ppf(self, probability: _X, *args: *_Xs) -> _Y: ...
    def median(self, *args: *_Xs) -> _Y: ...
    def isf(self, probability: _X, *args: *_Xs) -> _Y: ...
    def ichf(self, cumulative_hazard_rate: _X, *args: *_Xs) -> _Y: ...
    def moment(self, n: int, *args: *_Xs) -> _Y: ...
    def mean(self, *args: *_Xs) -> _Y: ...
    def var(self, *args: *_Xs) -> _Y: ...
    def mrl(self, time: _X, *args: *_Xs) -> _Y: ...
    def ls_integrate(self, func: Callable[[_X], _Y], a: _X, b: _X, *args: *_Xs, deg: int = 10) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B]: ...
    @overload
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _Y]: ...
    @overload
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B, _Y]: ...
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> _Y | tuple[_Y, _Y] | tuple[_Y, _B] | tuple[_Y, _B, _Y]: ...
    @property
    def plot(self) -> PlotParametricLifetimeModel: ...

class FittableParametricLifetimeModel(ParametricLifetimeModel[*_Xs], ABC):
    fitting_results = Optional[FittingResults]

    def __init__(self, **kwparams: Optional[float]) -> None: ...
    @abstractmethod
    def dhf(self, time: _X, *args: *_Xs) -> _Y: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    @abstractmethod
    def jac_hf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    @abstractmethod
    def jac_chf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_sf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_sf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    @abstractmethod
    def jac_sf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_pdf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: Literal[False] = False,
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_pdf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: Literal[True] = True,
    ) -> _Y: ...
    @abstractmethod
    def jac_pdf(
        self,
        time: _X,
        *args: *_Xs,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @abstractmethod
    def _get_params_bounds(self) -> Bounds: ...
    @abstractmethod
    def _get_initial_params(
        self,
        time: NDArray[np.float64],
        *args: *_Xs,
        event: Optional[NDArray[np.bool_]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
    ) -> NDArray[np.float64]: ...
    def fit(
        self,
        time: NDArray[np.float64],
        *args: *_Xs,
        event: Optional[NDArray[np.bool_]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
        **options: Any,
    ) -> None: ...

class NonParametricLifetimeModel(ABC):
    @abstractmethod
    def fit(
        self,
        time: NDArray[np.float64],
        event: Optional[NDArray[np.bool_]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
    ) -> Self: ...

class FrozenParametricLifetimeModel(FrozenParametricModel[*_Xs]):

    unfrozen_model: ParametricLifetimeModel[*_Xs]

    def __init__(self, model: ParametricLifetimeModel[*_Xs], *args: *_Xs) -> None: ...
    def hf(self, time: _X) -> _Y: ...
    def chf(self, time: _X) -> _Y: ...
    def sf(self, time: _X) -> _Y: ...
    def pdf(self, time: _X) -> _Y: ...
    def mrl(self, time: _X) -> _Y: ...
    def moment(self, n: int) -> _Y: ...
    def mean(self) -> _Y: ...
    def var(self) -> _Y: ...
    def isf(self, probability: _X) -> _Y: ...
    def ichf(self, cumulative_hazard_rate: _X) -> _Y: ...
    def cdf(self, time: _X) -> _Y: ...
    def ppf(self, probability: _X) -> _Y: ...
    def median(self) -> _Y: ...
    def ls_integrate(
        self,
        func: Callable[[_X], _Y],
        a: _X,
        b: _X,
        deg: int = 10,
    ) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B]: ...
    @overload
    def rvs(
        self,
        size: int,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _Y]: ...
    @overload
    def rvs(
        self,
        size: int,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B, _Y]: ...
    def rvs(
        self,
        size: int,
        *args: *_Xs,
        nb_assets: Optional[int] = None,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> _Y | tuple[_Y, _Y] | tuple[_Y, _B] | tuple[_Y, _B, _Y]: ...



import numpy as np

from relife.data import LifetimeData
from relife.likelihood import LikelihoodFromLifetimes


class TestLikelihoodDistribution:
    def test_negative_log(self, distribution, power_transformer_data):
        lifetime_data = LifetimeData(
            power_transformer_data["time"], event=power_transformer_data["event"], entry=power_transformer_data["entry"]
        )
        likelihood = LikelihoodFromLifetimes(distribution, lifetime_data)
        assert likelihood.negative_log(distribution.params).shape == ()

    def test_jac_negative_log(self, distribution, power_transformer_data):
        lifetime_data = LifetimeData(
            power_transformer_data["time"], event=power_transformer_data["event"], entry=power_transformer_data["entry"]
        )
        likelihood = LikelihoodFromLifetimes(distribution, lifetime_data)
        assert likelihood.jac_negative_log(distribution.params).shape == (distribution.nb_params,)


class TestLikelihoodRegression:
    def test_negative_log(self, regression, insulator_string_data):
        covar = np.column_stack(
            (
                insulator_string_data["pHCl"],
                insulator_string_data["pH2SO4"],
            )
        )
        lifetime_data = LifetimeData(
            insulator_string_data["time"], insulator_string_data["event"], insulator_string_data["entry"], args=(covar,)
        )
        likelihood = LikelihoodFromLifetimes(regression, lifetime_data)
        assert likelihood.negative_log(regression.params).shape == ()

    def test_jac_negative_log(self, regression, insulator_string_data):
        covar = np.column_stack(
            (
                insulator_string_data["pHCl"],
                insulator_string_data["pH2SO4"],
            )
        )
        lifetime_data = LifetimeData(
            insulator_string_data["time"], insulator_string_data["event"], insulator_string_data["entry"], args=(covar,)
        )
        likelihood = LikelihoodFromLifetimes(regression, lifetime_data)
        assert likelihood.jac_negative_log(regression.params).shape == (regression.nb_params,)



from relife.lifetime_model.distribution import (
    LifetimeDistribution as LifetimeDistribution,
)
from relife.lifetime_model.regression import CovarEffect as CovarEffect
from relife.lifetime_model.regression import (
    FrozenLifetimeRegression as FrozenLifetimeRegression,
)
from relife.lifetime_model.regression import LifetimeRegression as LifetimeRegression

from ._base import FittableParametricLifetimeModel as FittableParametricLifetimeModel
from ._base import FrozenParametricLifetimeModel as FrozenParametricLifetimeModel
from ._base import NonParametricLifetimeModel as NonParametricLifetimeModel
from ._base import ParametricLifetimeModel as ParametricLifetimeModel
from .conditional_model import AgeReplacementModel as AgeReplacementModel
from .conditional_model import FrozenAgeReplacementModel as FrozenAgeReplacementModel
from .conditional_model import FrozenLeftTruncatedModel as FrozenLeftTruncatedModel
from .conditional_model import LeftTruncatedModel as LeftTruncatedModel
from .distribution import EquilibriumDistribution as EquilibriumDistribution
from .distribution import Exponential as Exponential
from .distribution import Gamma as Gamma
from .distribution import Gompertz as Gompertz
from .distribution import LogLogistic as LogLogistic
from .distribution import MinimumDistribution as MinimumDistribution
from .distribution import Weibull as Weibull
from .non_parametric import ECDF as ECDF
from .non_parametric import KaplanMeier as KaplanMeier
from .non_parametric import NelsonAalen as NelsonAalen
from .non_parametric import Turnbull as Turnbull
from .regression import AcceleratedFailureTime as AcceleratedFailureTime
from .regression import ProportionalHazard as ProportionalHazard



TIMELINE_DOCSTRING = """
Calculate the {name}  over a given timeline.

It takes into account ``discounting_rate`` attribute value.

{formula}

Parameters
----------
timeline: np.ndarray
    Values of the timeline over which the {name} is to be calculated.

Returns
-------
np.ndarray
    The {name}.
"""

ASYMPTOTIC_DOCSTRING = """
Calculate the asymptotic {name}.

It takes into account ``discounting_rate`` attribute value.

{formula}

Returns
-------
np.ndarray
    The asymptotic {name}.
"""

ETC_FORMULA = r"""
The expected total cost :math:`z(t)` is computed by solving the renewal equation and is given by:

.. math::

    z(t) = \mathbb{E}(Z_t) = \int_{0}^{\infty}\mathbb{E}(Z_t~|~X_1 = x)dF(x)

where :

- :math:`t` is the time
- :math:`X_i \sim F` are :math:`n` random variable lifetimes, *i.i.d.*, of cumulative distribution :math:`F`.
- :math:`Z_t` is the random variable reward at each time :math:`t`.
- :math:`\delta` is the discounting rate.
"""

EEAC_FORMULA = r"""
The expected equivalent annual cost :math:`\text{EEAC}(t)` is given by:

.. math::

    \text{EEAC}(t) = \dfrac{\delta z(t)}{1 - e^{-\delta t}}

where :

- :math:`t` is the time
- :math:`z(t)` is the expected_total_cost at :math:`t`. See :py:meth:`~AgeReplacementPolicy.expected_total_cost` for more details.`.
- :math:`\delta` is the discounting rate.
"""

ASYMPTOTIC_ETC_FORMULA = r"""
The asymptotic expected total cost is:
 
.. math::

    \lim_{t\to\infty} z(t)

where :math:`z(t)` is the expected total cost at :math:`t`. See :py:meth:`~AgeReplacementPolicy.expected_total_cost` for more details.
"""

ASYMPTOTIC_EEAC_FORMULA = r"""
The asymptotic expected total cost is:

.. math::

    \lim_{t\to\infty} \text{EEAC}(t)
    
where :math:`\text{EEAC}(t)` is the expected equivalent annual cost at :math:`t`. See :py:meth:`~AgeReplacementPolicy.expected_equivalent_annual_cost` for more details.
"""

ETC_DOCSTRING = TIMELINE_DOCSTRING.format(name="expected total cost", formula=ETC_FORMULA)
EEAC_DOCSTRING = TIMELINE_DOCSTRING.format(name="expected equivalent annual cost", formula=EEAC_FORMULA)
ASYMPTOTIC_ETC_DOCSTRING = ASYMPTOTIC_DOCSTRING.format(name="expected total cost", formula=ASYMPTOTIC_ETC_FORMULA)
ASYMPTOTIC_EEAC_DOCSTRING = ASYMPTOTIC_DOCSTRING.format(
    name="expected equivalent annual cost", formula=ASYMPTOTIC_EEAC_FORMULA
)
